import argparse
import time
from typing import Optional

from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock

import socs.agents.hwp_gripper.drivers.gripper_client as cli
from socs.agents.hwp_supervisor.agent import get_op_data


class HWPGripperAgent:
    """
    Agent for controlling/monitoring the HWP's three LEY32C-30 linear actuators.
    This interfaces with the GripperServer running on the beaglebone
    microcontroller (https://github.com/simonsobs/sobonelib/blob/main/hwp_gripper/control/gripper_server.py).

    This agent will issue commands to the microcontroller via OCS, and publish
    gripper position and limit switch states to grafana.

    Args:
        agent (ocs.ocs_agent.OCSAgent):
            Agent instance
        args (argparse.Namespace):
            Parsed command line arguments namespace.
    """

    def __init__(self, agent, args):
        self.agent = agent
        self.log = agent.log
        self.client_lock = TimeoutLock()

        self._initialized = False
        self.mcu_ip = args.mcu_ip
        self.control_port = args.control_port
        self.warm_grip_distance = args.warm_grip_distance
        self.adjustment_distance = args.adjustment_distance

        self.shutdown_mode = False
        self.supervisor_id = args.supervisor_id

        self.client: Optional[cli.GripperClient] = None

        self.decoded_alarm_group = {False: 'No alarm was detected',
                                    'A': 'Unrecognized error',
                                    'B': 'Controller saved parameter issue',
                                    'C': 'Issue with position calibration',
                                    'D': 'Could not reach position within time limit',
                                    'E': 'General controller erorr'}

        agg_params = {'frame_length': 60}
        self.agent.register_feed('hwp_gripper', record=True, agg_params=agg_params)
        self.agent.register_feed('gripper_action', record=True)

    def _run_client_func(self, func, *args, lock_timeout=10,
                         job=None, check_shutdown=True, **kwargs):
        """
        Args
        ----
        func (function):
            Function to run
        *args (positional args):
            Additional args to pass to function
        lock_timeout (float):
            Time (sec) to wait to acquire client-lock before throwing an error.
        job (str):
            Name of job to attach to lock
        check_shutdown (bool):
            If true, will block the client function and throw an error if
            shutdown mode is in effect.
        """
        if self.shutdown_mode and check_shutdown:
            raise RuntimeError(
                'Cannot run client function, shutdown mode is in effect'
            )

        lock_kw = {'timeout': lock_timeout}
        if job is not None:
            lock_kw['job'] = job
        with self.client_lock.acquire_timeout(**lock_kw) as acquired:
            if not acquired:
                self.log.error(
                    f"Could not acquire lock! Job {self.client_lock.job} is "
                    "already running."
                )
                raise TimeoutError('Could not acquire lock')

            return_dict = func(*args, **kwargs)

        for line in return_dict['log']:
            self.log.debug(line)
        return return_dict

    def _get_hwp_freq(self):
        if self.supervisor_id is None:
            raise ValueError("No Supervisor ID set")

        res = get_op_data(self.supervisor_id, 'monitor')
        return res['data']['hwp_state']['pid_current_freq']

    def _check_stopped(self):
        return self._get_hwp_freq() < 0.1

    def init_connection(self, session, params):
        """init_connection()

        **Task** - Initialize connection to the GripperServer on the BeagleBone
        micro-controller
        """
        if self._initialized:
            self.log.info('Connection already initialized. Returning...')
            return True, 'Connection already initialized'

        self.client = cli.GripperClient(self.mcu_ip, self.control_port)
        self.log.info("Initialized Client")
        self._initialized = True
        return True, 'Initialized connection to GripperServer'

    @ocs_agent.param('state', default=True, type=bool)
    def power(self, session, params=None):
        """power(state=True)

        **Task** - If turning on, will power on the linear actuators and disengage
        brakes. If turning off, will cut power to the linear actuators and engage
        brakes.

        Parameters:
            state (bool): State to set the actuator power to.

        Notes:
            The most recent data collected is stored in session data in the
            structure:

                >>> response.session['response']
                {'result': True,
                 'log': ["SVON turned on in Control.ON()",
                         "Turned off BRAKE for axis 1 in Control.BRAKE()",
                         "Turned off BRAKE for axis 2 in Control.BRAKE()",
                         "Turned off BRAKE for axis 3 in Control.BRAKE()",
                         "Successfully turned off BRAKE for axis 1 in Control.BRAKE()",
                         "Successfully turned off BRAKE for axis 2 in Control.BRAKE()",
                         "Successfully turned off BRAKE for axis 3 in Control.BRAKE()",
                         "Disengaged brakes in Control.ON()"]}
        """
        return_dict = self._run_client_func(
            self.client.power, params['state'], job='power'
        )
        session.data['response'] = return_dict
        return return_dict['result'], f"Success: {return_dict['result']}"

    @ocs_agent.param('state', default=True, type=bool)
    @ocs_agent.param('actuator', default=0, type=int, check=lambda x: 0 <= x <= 3)
    def brake(self, session, params=None):
        """brake(state=True, actuator=0)

        **Task** - Controls actuator brakes

        Parameters:
            state (bool):
                State to set the actuator brake to. Takes bool input
            actuator (int):
                Actuator number. Takes input of 0-3 with 1-3 controlling and
                individual actuator and 0 controlling all three

        Notes:
            The most recent data collected is stored in session data in the
            structure:

                >>> response.session['response']
                {'result': True,
                 'log': ["Turned off BRAKE for axis 1 in Control.BRAKE()",
                         "Successfully turned off BRAKE for axis 1 in Control.BRAKE()"]}
        """
        return_dict = self._run_client_func(
            self.client.brake, params['state'], params['actuator'], job='brake'
        )
        session.data['response'] = return_dict
        return return_dict['result'], f"Success: {return_dict['result']}"

    @ocs_agent.param('mode', default='push', type=str, choices=['push', 'pos'])
    @ocs_agent.param('actuator', default=1, type=int, check=lambda x: 1 <= x <= 3)
    @ocs_agent.param('distance', default=0, type=float)
    def move(self, session, params=None):
        """move(mode='push', actuator=1, distance=0)

        **Task** - Move an actuator a specific distance. If the HWP is spinning,
        this task will not run.

        Parameters:
            mode (str):
                Movement mode. Takes inputs of 'pos' (positioning) or 'push'
                (pushing)
            actuator (int):
                Actuator number 1-3
            distance (float):
                Distance to move (mm). Takes positive and negative numbers for
                'pos' mode. Takes only positive numbers for 'push' mode. Value
                should be a multiple of 0.1.

        Notes:
            Positioning mode is used when you want to position the actuators without
            gripping the rotor. Pushing mode is used when you want the grip the
            rotor.

            The most recent data collected is stored in session data in the
            structure:

                >>> response.session['response']
                {'result': True,
                 'log': ["Control.STEP() operation finished for step 1",
                         "MOVE in Gripper.MOVE() completed successfully"]}
        """

        if self._get_hwp_freq() > 0.1:
            self.log.warn("Not moving actuators while HWP is spinning")
            return False, "HWP is spinning, not moving actuators"

        return_dict = self._run_client_func(
            self.client.move, params['mode'], params['actuator'],
            params['distance'], job='move'
        )
        session.data['response'] = return_dict
        return return_dict['result'], f"Success: {return_dict['result']}"

    def home(self, session, params=None):
        """home()
        **Task** - Homes and recalibrates the position of the actuators

        Note:
            This action must be done first after a power cycle. Otherwise the
            controller will throw an error.

            The most recent data collected is stored in session data in the
            structure:

                >>> response.session['response']
                {'result': True,
                 'log': ["SVON turned on in Control.ON()",
                         "'HOME' operation finished in Control.HOME()",
                         "HOME operation in Gripper.HOME() completed"]}
        """
        return_dict = self._run_client_func(self.client.home, job='home')
        session.data['response'] = return_dict
        return return_dict['result'], f"Success: {return_dict['result']}"

    def inp(self, session, params=None):
        """inp()

        **Task** - Queries whether the actuators are in a known position. This
        tells you whether the windows software has detected that the actuator
        has been homed.

        Notes:
            The most recent data collected is stored in session data in the
            structure:

                >>> response.session['response']
                {'result': True,
                 'log': []}
        """
        return_dict = self._run_client_func(
            self.client.inp, job='INP', check_shutdown=False)
        session.data['response'] = return_dict
        return return_dict['result'], f"Success: {return_dict['result']}"

    def alarm(self, session, params=None):
        """alarm()

        **Task** - Queries the actuator controller alarm state

        Notes:
            The most recent data collected is stored in session data in the
            structure:

                >>> response.session['response']
                {'result': True,
                 'log': ["ALARM = False"]}
        """
        return_dict = self._run_client_func(
            self.client.alarm, job='alarm', check_shutdown=False)
        session.data['response'] = return_dict
        return return_dict['result'], f"Success: {return_dict['result']}"

    def alarm_group(self, session, params=None):
        """alarm_group()

        **Task** - Queries the actuator controller alarm group

        Notes:
            Return one of six values depending on the controller alarm state
                False: No alarm was detected
                'A': Unrecognized error
                'B': Controller saved parameter issue
                'C': Issue with position calibration
                'D': Could not reach position within time limit
                'E': General controller error

            The most recent data collected is stored in session data in the
            structure:

                >>> response.session['response']
                {'result': 'B',
                 'log': ["ALARM GROUP B detected"]}
        """
        state_return_dict = self._run_client_func(
            self.client.get_state, job='get_state', check_shutdown=False)

        if not bool(state_return_dict['result']['jxc']['alarm']):
            return_dict = {'result': False, 'log': ['Alarm not triggered']}
        else:
            return_dict = self._run_client_func(
                self.client.alarm_group, job='alarm_group', check_shutdown=False)

            if return_dict['result'] is None:
                return_dict['result'] = 'A'

        session.data['response'] = return_dict
        session.data['decoded_alarm_group'] = self.decoded_alarm_group[return_dict['result']]
        return return_dict['result'], f"Success: {return_dict['result']}"

    def reset(self, session, params=None):
        """reset()
        **Task** - Resets the current active controller alarm

        Notes:
            The most recent data collected is stored in session data in the
            structure:

                >>> response.session['response']
                {'result': True,
                 'log': ["Ignored Control.ALARM_GROUP(). No ALARM detected",
                         "RESET aborted in Gripper.RESET() due to no detected alarm"]}
        """
        return_dict = self._run_client_func(self.client.reset, job='reset')
        session.data['response'] = return_dict
        return return_dict['result'], f"Success: {return_dict['result']}"

    @ocs_agent.param('actuator', default=1, type=int, check=lambda x: 1 <= x <= 3)
    def act(self, session, params=None):
        """act(actuator=1)

        **Task** - Queries whether an actuator is connected

        Parameters:
            actuator (int): Actuator number 1-3

        Notes:
            The most recent data collected is stored in session data in the
            structure:

                >>> response.session['response']
                {'result': True,
                 'log': ["Actuator 1 is in state 1"]}
        """
        return_dict = self._run_client_func(
            self.client.act, params['actuator'], job='act', check_shutdown=False)
        session.data['response'] = return_dict
        return return_dict['result'], f"Success: {return_dict['result']}"

    @ocs_agent.param('value', type=bool)
    def is_cold(self, session, params=None):
        """is_cold(value=False)

        **Task** - Set the limit switches to warm/cold grip configuration

        Parameters:
            value (bool): Set to warm grip (False) or cold grip (True)

        Notes:
            Configures the software to query the correct set of limit switches.
            Warm grip configuration enables both warm and cold limit switches.
            Cold grip configuration enables only cold limit switches. The maximum
            extension of the actuators depends on the cryostat temperature.

            The most recent data collected is stored in session data in the
            structure:

                >>> response.session['response']
                {'result': True,
                 'log': ["Received request to change is_cold to True",
                         "is_cold successfully changed"]}
        """
        return_dict = self._run_client_func(
            self.client.is_cold, params['value'], job='is_cold')
        session.data['response'] = return_dict
        return return_dict['result'], f"Success: {return_dict['result']}"

    @ocs_agent.param('value', default=False, type=bool)
    def force(self, session, params=None):
        """force(value=False)

        **Task** - Enable or disable force mode in the GripperServer,
        which will ignore limit switch information.

        Parameters:
            value (bool): Use limit switch information (False) or ignore limit
                switch information (True)

        Notes:
            By default the code is configured to prevent actuator movement if
            on of the limit switches has been triggered. This function can be
            called to forcibly move the actuators even with a limit switch
            trigger.

            The most recent data collected is stored in session data in the
            structure:

                >>> response.session['response']
                {'result': True,
                 'log': ["Received request to change force to True",
                         "force successfully changed"]}
        """
        return_dict = self._run_client_func(
            self.client.force, params['value'], job='force')
        session.data['response'] = return_dict
        return return_dict['result'], f"Success: {return_dict['result']}"

    def shutdown(self, session, params=None):
        """shutdown()

        **Task** - Series of commands executed during a shutdown.
        This will enable shutdown mode, which locks out other agent operations.

        Notes:
            This function is called once a shutdown trigger has been given.
        """
        self.log.info("Enabling shutdown-mode, which will prevent grip "
                      "operations from being performed")
        self.shutdown_mode = True
        return True, 'Shutdown completed'

    def grip(self, session, params=None):
        """grip()

        **Task** - Series of commands to automatically warm grip the HWP. This
        assumes that HWP is cold. This will return grippers to their home position,
        then move them each inwards incrementally until warm limit switches are
        tiggered. If the HWP is spinning, this will not run. If this fails to grip
        hwp, this will return grippers to their home position.

        Notes:
            The most recent data collected is stored in session data in the
            structure:

                >>> response.session['responses']
                [{'result': True, 'log': [.., .., etc]},
                                    ..
                 {'result': True, 'log': [.., .., etc]}]
        """
        if self._get_hwp_freq() > 0.1:
            return False, "Not gripping HWP because HWP is spinning"

        result, session.data = self._grip_hwp(check_shutdown=True)
        return result, "Finished gripping procedure"

    def _grip_hwp(self, check_shutdown=True):
        """
        Helper function to grip the HWP that can be called by the grip_hwp
        task or by the shutdown procedure.  This will return grippers to their
        home position, then move them each inwards incrementally.
        """
        data = {
            'responses': [],
        }

        def run_and_append(func, *args, **kwargs):
            return_dict = self._run_client_func(func, *args, **kwargs)
            data['responses'].append(return_dict)
            return return_dict

        # Check controller alarm
        return_dict = run_and_append(self.client.get_state, job='grip',
                                     check_shutdown=check_shutdown)
        if return_dict['result']['jxc']['alarm']:
            return_dict = self._run_client_func(
                self.client.alarm_group, job='alarm_group', check_shutdown=False)
            if return_dict['result'] is None:
                return_dict['result'] = 'A'
            alarm_message = self.decoded_alarm_group[return_dict['result']]
            self.log.error(
                f"Abort grip. Detected contoller alarm: {alarm_message}")
            return False, data

        # Check if hwp is already gripper or not
        act_results = return_dict['result']['actuators']
        limit_switch_state = act_results[0]['limits']['warm_grip']['state'] | \
            act_results[1]['limits']['warm_grip']['state'] | \
            act_results[2]['limits']['warm_grip']['state']
        if limit_switch_state:
            self.log.warn("HWP is already gripped. Do nothing.")
            return True, data

        # Reset alarms
        run_and_append(self.client.reset, job='grip', check_shutdown=check_shutdown)
        time.sleep(1)

        # Enable power to actuators
        run_and_append(self.client.power, True, job='grip', check_shutdown=check_shutdown)
        time.sleep(1)

        # Disable brake
        run_and_append(self.client.brake, False, job='grip', check_shutdown=check_shutdown)
        time.sleep(1)

        # Ignore limit switches to move grippers backwards
        run_and_append(self.client.force, True, job='grip', check_shutdown=check_shutdown)

        # Ensure that the limit switches are in warm configuration
        run_and_append(self.client.is_cold, False, job='grip', check_shutdown=check_shutdown)

        # Send grippers to their home position
        run_and_append(self.client.home, job='grip', check_shutdown=check_shutdown)

        # Ensure that we do not ignore limit switches
        run_and_append(self.client.force, False, job='grip', check_shutdown=check_shutdown)

        finished = [False, False, False]

        # Move to the warm_grip_distance - 5 mm.
        for actuator in range(3):
            distance = self.warm_grip_distance[actuator] - 5.
            return_dict = run_and_append(self.client.move, 'POS', actuator + 1, distance,
                                         job='grip', check_shutdown=check_shutdown)
            time.sleep(1)

        # Move actator inwards by 0.1 mm step, warm_grip_distance + 1 is absolute maximum
        for i in range(60):
            if all(finished):
                break
            for actuator, _ in enumerate(finished):
                if finished[actuator]:
                    continue

                # Move actuator inwards until warm-limit is hit
                # If the warm limit is hit, return_dict['result'] will be False
                return_dict = run_and_append(self.client.move, 'POS', actuator + 1, 0.1,
                                             job='grip', check_shutdown=check_shutdown)

                if not return_dict['result']:
                    # Reset alarms.
                    run_and_append(self.client.reset, job='grip',
                                   check_shutdown=check_shutdown)
                    time.sleep(1)

                    # If the warm-limit is hit, move the actuator outwards by 0.5 mm
                    # to un-trigger the warm-limit. This is because gripper sligthly
                    # overshoot the limit switches. The outward movement compensates
                    # for the overshoot and hysteresys of the limit switches.
                    run_and_append(self.client.is_cold, True, job='grip',
                                   check_shutdown=check_shutdown)
                    time.sleep(1)

                    run_and_append(self.client.move, 'POS', actuator + 1, -0.5,
                                   job='grip', check_shutdown=check_shutdown)

                    run_and_append(self.client.is_cold, False, job='grip',
                                   check_shutdown=check_shutdown)
                    time.sleep(1)

                    finished[actuator] = True

        # Return grippers back to home is something is wrong
        if not all(finished):
            self.log.error('Failed to grip HWP. Retract grippers.')
            run_and_append(self.client.force, True, job='grip',
                           check_shutdown=check_shutdown)
            time.sleep(1)

            run_and_append(self.client.home, job='grip',
                           check_shutdown=check_shutdown)
            time.sleep(1)

            run_and_append(self.client.force, False, job='grip',
                           check_shutdown=check_shutdown)

        # Adjust gripper position if you need
        else:
            run_and_append(self.client.is_cold, True, job='grip',
                           check_shutdown=check_shutdown)
            time.sleep(1)

            for actuator in range(3):
                run_and_append(self.client.move, 'POS', actuator + 1,
                               self.adjustment_distance[actuator],
                               job='grip', check_shutdown=check_shutdown)

            run_and_append(self.client.is_cold, False, job='grip',
                           check_shutdown=check_shutdown)
            time.sleep(1)

        # Enable breaks
        run_and_append(self.client.brake, True, job='grip',
                       check_shutdown=check_shutdown)
        time.sleep(1)

        # Disable power to actuators
        run_and_append(self.client.power, False, job='grip',
                       check_shutdown=check_shutdown)
        time.sleep(1)

        # We should stop schedule if we have an error in this task
        if not all(finished):
            self.log.error('Failed to grip HWP. Grippers are retracted.')
            return False, data

        return True, data

    def ungrip(self, session, params=None):
        """ungrip()

        **Task** - Series of commands to automatically ungrip the HWP.
        This will return grippers to their home position, and retract
        grippers as much as possible.

        Notes:
            The most recent data collected is stored in session data in the
            structure:

                >>> response.session['responses']
                [{'result': True, 'log': [.., .., etc]},
                                    ..
                 {'result': True, 'log': [.., .., etc]}]
        """

        result, session.data = self._ungrip_hwp(check_shutdown=True)
        return result, "Finished ungripping procedure"

    def _ungrip_hwp(self, check_shutdown=True):
        """
        Helper function to ungrip the HWP that can be called by the ungrip_hwp
        task or by the shutdown procedure.  This will return grippers to their
        home position, and retract as much as possible.
        """
        data = {
            'responses': [],
        }

        def run_and_append(func, *args, **kwargs):
            return_dict = self._run_client_func(func, *args, **kwargs)
            data['responses'].append(return_dict)
            return return_dict

        # Check controller alarm
        return_dict = run_and_append(self.client.get_state, job='ungrip',
                                     check_shutdown=check_shutdown)
        if return_dict['result']['jxc']['alarm']:
            return_dict = self._run_client_func(
                self.client.alarm_group, job='alarm_group', check_shutdown=False)
            if return_dict['result'] is None:
                return_dict['result'] = 'A'
            alarm_message = self.decoded_alarm_group[return_dict['result']]
            self.log.error(
                f"Abort ungrip. Detected contoller alarm: {alarm_message}")
            return False, data

        run_and_append(self.client.reset, job='ungrip', check_shutdown=check_shutdown)
        # Enable power to actuators
        run_and_append(self.client.power, True, job='ungrip', check_shutdown=check_shutdown)
        # Disable breaks
        run_and_append(self.client.brake, False, job='ungrip', check_shutdown=check_shutdown)
        # Ignore limit switches
        run_and_append(self.client.force, True, job='ungrip', check_shutdown=check_shutdown)

        # Send grippers to their home position
        run_and_append(self.client.home, job='ungrip', check_shutdown=check_shutdown)

        # Move actuator outwards as much as possible
        for actuator in range(1, 4):
            run_and_append(self.client.move, 'POS', actuator, -1.9,
                           job='ungrip', check_shutdown=check_shutdown)
        time.sleep(1)

        # Enable brake
        run_and_append(self.client.brake, True, job='ungrip', check_shutdown=check_shutdown)
        # Power off actuators
        run_and_append(self.client.power, False, job='ungrip', check_shutdown=check_shutdown)
        # Enable limit switches
        run_and_append(self.client.force, False, job='ungrip', check_shutdown=check_shutdown)
        time.sleep(1)

        # check limit switch state
        return_dict = run_and_append(self.client.get_state, job='ungrip',
                                     check_shutdown=check_shutdown)
        act_results = return_dict['result']['actuators']
        limit_switch_state = act_results[0]['limits']['warm_grip']['state'] | \
            act_results[1]['limits']['warm_grip']['state'] | \
            act_results[2]['limits']['warm_grip']['state']

        # Stop schedule if the limit switch state is wrong
        if limit_switch_state:
            self.log.error("Failed to ungrip HWP. Limit switch state is not as expected.")
            return False, data

        return True, data

    def cancel_shutdown(self, session, params=None):
        """cancel_shutdown()

        **Task** - Take the gripper agent out of shutdown mode
        """
        self.shutdown_mode = False
        return True, 'Cancelled shutdown mode'

    def restart(self, session, params=None):
        """restart()

        **Task** - Restarts the beaglebone processes and the socket connection

        Notes:
            The most recent data collected is stored in session data in the
            structure:

                >>> response.session['response']
                {'result': True,
                 'log': ["Previous connection closed",
                         "Restart command response: Success",
                         "Control socket reconnected"]}
        """
        return_dict = self._run_client_func(
            self.client.restart, job='restart', check_shutdown=False)
        session.data['response'] = return_dict
        return return_dict['result'], f"Success: {return_dict['result']}"

    def monitor_state(self, session, params=None):
        """monitor_state()

        **Process** - Process to monitor the gripper state

        Notes:
            The most recent data collected is stored in session data in the
            structure:

                >>> response.session
                {'last_updated': 1649085992.719602,
                 'state': {'last_packet_received': 1649085992.719602,
                           'jxc_setup': False,
                           'jxc_svon': False,
                           'jxc_busy': False,
                           'jxc_seton': False,
                           'jxc_inp': False,
                           'jxc_svre': False,
                           'jxc_alarm': False,
                           'jxc_alarm_message': 'No alarm was detected',
                           'jxc_out': 0,
                           'act{axis}_pos': 0,
                           'act{axis}_limit_warm_grip_state': False,
                           'act{axis}_limit_cold_grip_state': False,
                           'act{axis}_emg': False,
                           'act{axis}_brake': True}}
        """
        sleep_time = 5
        while session.status in ['starting', 'running']:
            if self.client is None:
                self.log.warn("Client not initialized")
                time.sleep(1)
                continue

            try:
                return_dict = self._run_client_func(
                    self.client.get_state, job='get_state', check_shutdown=False
                )
            except TimeoutError:
                self.log.warn('monitor_state: Query Timeout')

            now = time.time()

            # Dict of the 'GripperState' class from the pru_monitor
            state = return_dict['result']
            data = {
                'last_packet_received': state['last_packet_received'],
            }

            data.update({
                'jxc_setup': int(state['jxc']['setup']),
                'jxc_svon': int(state['jxc']['svon']),
                'jxc_busy': int(state['jxc']['busy']),
                'jxc_seton': int(state['jxc']['seton']),
                'jxc_inp': int(state['jxc']['inp']),
                'jxc_svre': int(state['jxc']['svre']),
                'jxc_alarm': int(state['jxc']['alarm']),
                'jxc_out': int(state['jxc']['out']),
            })

            alarm_group_mapping = {4: 'B',
                                   2: 'C',
                                   1: 'D',
                                   0: 'E'}

            if bool(state['jxc']['alarm']):
                out_value = int(state['jxc']['out']) % 16
                if out_value in alarm_group_mapping.keys():
                    alarm_message = self.decoded_alarm_group[alarm_group_mapping[out_value]]
                else:
                    alarm_message = self.decoded_alarm_group['A']
            else:
                alarm_message = self.decoded_alarm_group[False]

            data.update({'jxc_alarm_message': alarm_message})

            for act in state['actuators']:
                axis = act['axis']
                data.update({
                    f'act{axis}_pos': act['pos'],
                    f'act{axis}_limit_cold_grip_state': int(act['limits']['cold_grip']['state']),
                    f'act{axis}_limit_warm_grip_state': int(act['limits']['warm_grip']['state']),
                    f'act{axis}_brake': int(act['brake']),
                    f'act{axis}_emg': int(act['emg']),
                })

            session.data = {
                'state': data,
                'last_updated': now,
            }

            _data = {
                'block_name': 'gripper_state',
                'timestamp': now,
                'data': data,
            }

            self.agent.publish_to_feed('hwp_gripper', _data)
            time.sleep(sleep_time)

    def _stop_monitor_state(self, session, params=None):
        session.set_status('stopping')
        return True, "Requesting monitor_state process to stop"

    @ocs_agent.param('no_data_warn_time', default=60, type=float)
    @ocs_agent.param('no_data_shutdown_time', default=300, type=float)
    def monitor_supervisor(self, session, params=None):
        """monitor_supervisor(no_data_warn_time=60, no_data_shutdown_time=300)

        **Process** - Monitor the hwp-supervisor state. If supervisor sends shutdown
        signal, or if communication wtih supervisor is dropped for longer than
        ``no_data_shutdown_time``, this will begin agent shutdown and disable access
        to other dangerous gripper operations.

        Parameters:
            no_data_warn_time (int): Time in seconds to wait after communication failure
                                     before generating a warning
            no_data_shutdown_time (int): Time in seconds to wait after communication failure
                                         before initiating shutdown

        Notes:
            The most recent data collected is stored in session data in the
            structure:

                >>> response.session
                {'time': 1649085992.719602,
                 'gripper_action': 'ok'}
        """
        last_ok_time = time.time()

        if self.supervisor_id is None:
            return False, 'No supervisor ID set'

        warning_issued = False
        while session.status in ['starting', 'running']:
            res = get_op_data(self.supervisor_id, 'monitor')
            if res['status'] != 'ok':
                action = 'no_data'
            else:
                action = res['data']['actions']['gripper']

            if action == 'ok':
                warning_issued = False
                last_ok_time = time.time()
            elif action == 'stop':
                if not self.shutdown_mode:
                    self.agent.start('shutdown')

            time_since_ok = time.time() - last_ok_time
            if time_since_ok > params['no_data_warn_time'] and not warning_issued:
                self.log.error(
                    f"Have not received 'ok' in {time_since_ok / 60:.2f} minutes."
                    f"Will issue shutdown in "
                    f"{params['no_data_shutdown_time'] / 60:.2f} minutes."
                )
                warning_issued = True

            if time_since_ok > params['no_data_shutdown_time']:
                self.log.error(
                    f"Have not received 'ok' in "
                    f"{params['no_data_shutdown_time'] / 60:.2f} minutes. "
                    "Issuing shutdown"
                )
                self.agent.start('shutdown')

            data = {
                'data': {'gripper_action': action},
                'block_name': 'gripper_action',
                'timestamp': time.time()
            }

            self.agent.publish_to_feed('gripper_action', data)
            session.data = {
                'gripper_action': action,
                'time': time.time()
            }

            time.sleep(1)
        return True, 'Gripper monitor exited cleanly'

    def _stop_monitor_supervisor(self, session, params=None):
        session.set_status('stopping')
        return True, "Requesting monitor_supervisor process to stop"


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--mcu-ip', type=str,
                        help='IP of Gripper Beaglebone')
    pgroup.add_argument('--control-port', type=int, default=8041,
                        help='Port for actuator control as set by the Beaglebone code')
    pgroup.add_argument('--warm-grip-distance', action='store', type=float, nargs=3,
                        default=[10.0, 10.0, 10.0],
                        help='Nominal distance for warm grip position (mm). This needs'
                             'to be multiple of 0.1')
    pgroup.add_argument('--adjustment-distance', action='store', type=float, nargs=3,
                        default=[0, 0, 0],
                        help='Adjustment distance to compensate the misalignment of '
                             'limit switches (mm). This needs to be multiple of 0.1')
    pgroup.add_argument('--supervisor-id', type=str,
                        help='Instance ID for HWP Supervisor agent')
    pgroup.add_argument('--no-data-warn-time', type=float, default=60,
                        help='Time (seconds) since last supervisor-ok signal to '
                             'wait before issuing a warning')
    pgroup.add_argument('--no-data-shutdown-time', type=float, default=300,
                        help='Time (seconds) since last supervisor-ok signal to '
                             'wait before entering shutdown mode')
    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='HWPGripperAgent',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)
    gripper_agent = HWPGripperAgent(agent, args)
    agent.register_task('init_connection', gripper_agent.init_connection,
                        startup=True)
    agent.register_process('monitor_state', gripper_agent.monitor_state,
                           gripper_agent._stop_monitor_state, startup=True)
    agent.register_process('monitor_supervisor', gripper_agent.monitor_supervisor,
                           gripper_agent._stop_monitor_supervisor,
                           startup={'no_data_warn_time': args.no_data_warn_time,
                                    'no_data_shutdown_time': args.no_data_shutdown_time})
    agent.register_task('power', gripper_agent.power)
    agent.register_task('brake', gripper_agent.brake)
    agent.register_task('move', gripper_agent.move)
    agent.register_task('home', gripper_agent.home)
    agent.register_task('inp', gripper_agent.inp)
    agent.register_task('alarm', gripper_agent.alarm)
    agent.register_task('alarm_group', gripper_agent.alarm_group)
    agent.register_task('reset', gripper_agent.reset)
    agent.register_task('act', gripper_agent.act)
    agent.register_task('is_cold', gripper_agent.is_cold)
    agent.register_task('force', gripper_agent.force)
    agent.register_task('shutdown', gripper_agent.shutdown)
    agent.register_task('grip', gripper_agent.grip)
    agent.register_task('ungrip', gripper_agent.ungrip)
    agent.register_task('cancel_shutdown', gripper_agent.cancel_shutdown)
    agent.register_task('restart', gripper_agent.restart)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import argparse
import os
import time

import txaio
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import Pacemaker, TimeoutLock

from socs.agents.kikusui_pcr500ma.drivers import PCR500MA, VOLT_ULIM_SOFT

PORT_DEFAULT = 5025

LOCK_RELEASE_SEC = 1.
LOCK_RELEASE_TIMEOUT = 10
ACQ_TIMEOUT = 5


class PCR500MAAgent:
    """OCS agent class for PCR500MA current source

    Parameters
    ----------
    ip : string
        IP address
    port : int
        Port number
    """

    def __init__(self, agent, ip_addr, port=PORT_DEFAULT):
        self.active = True
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()
        self.take_data = False

        self._dev = PCR500MA(ip_addr=ip_addr, port=port)

        self.initialized = False

        agg_params = {'frame_length': 60}
        self.agent.register_feed('heater_source',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=1)

    @ocs_agent.param('sampling_frequency', default=0.2, type=float)
    def acq(self, session, params):
        """acq(sampling_frequency=0.2)

        **Process** - Monitor status of the relay.

        Parameters
        ----------
        sampling_frequency : float, optional
            Sampling frequency in Hz, defaults to 0.2 Hz.

        Notes
        -----
        An example of the session data::

            >>> response.session['data']
            {"I_AC": 0.00475843,
             "V_AC": 0.0371215,
             "P_AC": 0.0,
             "f_AC": 60.0,
             "output": 0,
             "timestamp": 1737367649.2595236
            }
        """
        f_sample = params.get('sampling_frequency')
        pace_maker = Pacemaker(f_sample)

        with self.lock.acquire_timeout(timeout=0, job='acq') as acquired:
            if not acquired:
                self.log.warn(
                    f'Could not start acq because {self.lock.job} is already running')
                return False, 'Could not acquire lock.'

            self.take_data = True
            last_release = time.time()

            while self.take_data:
                # Release lock
                if time.time() - last_release > LOCK_RELEASE_SEC:
                    last_release = time.time()
                    if not self.lock.release_and_acquire(timeout=LOCK_RELEASE_TIMEOUT):
                        print(f'Re-acquire failed: {self.lock.job}')
                        return False, 'Could not re-acquire lock.'

                # Data acquisition
                current_time = time.time()
                data = {'timestamp': current_time, 'block_name': 'heater_source', 'data': {}}

                try:
                    i_ac = self._dev.meas_current_ac()
                    v_ac = self._dev.meas_volt_ac()
                    p_ac = self._dev.meas_power_ac()
                    f_ac = self._dev.meas_freq()
                    sw_status = self._dev.get_output()
                    if session.degraded:
                        self.log.info('Connection re-established.')
                        session.degraded = False
                except ConnectionError:
                    self.log.error('Failed to get data from Kikusui PCR. Check network connection.')
                    session.degraded = True
                    time.sleep(1)
                    continue

                data['data']['I_AC'] = i_ac
                data['data']['V_AC'] = v_ac
                data['data']['P_AC'] = p_ac
                data['data']['f_AC'] = f_ac
                data['data']['output'] = 1 if sw_status else 0

                field_dict = {'I_AC': i_ac,
                              'V_AC': v_ac,
                              'P_AC': p_ac,
                              'f_AC': f_ac,
                              'output': 1 if sw_status else 0}

                session.data.update(field_dict)

                self.agent.publish_to_feed('heater_source', data)
                session.data.update({'timestamp': current_time})

                pace_maker.sleep()

            self.agent.feeds['heater_source'].flush_buffer()

        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params=None):
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data.'

        return False, 'acq is not currently running.'

    @ocs_agent.param('output', type=bool)
    @ocs_agent.param('force', default=False, type=bool)
    def set_output(self, session, params=None):
        """set_output(output, force=False)

        **Task** - Turns the output on or off.

        Parameters
        ----------
        output : bool
            True : output on
            False : output off

        force : bool, default False
            Force output on / off without checking output value.

        Notes
        -----
        This function measures the output voltage when turning off
        and checks the voltage setting when turning on by default. If the
        voltage is too high it will not change state to avoid an abrupt
        change in applied voltage.
        Set `force` option to `True` when this behavior is unwanted.
        """
        with self.lock.acquire_timeout(ACQ_TIMEOUT, job='set_output') as acquired:
            if not acquired:
                self.log.warn('Could not start set_output because '
                              f'{self.lock.job} is already running')
                return False, 'Could not acquire lock.'

            output_on = params['output']
            force = params['force']
            # Use measurement value when checking turning off threshold.
            use_meas_val = not output_on

            if force:
                self._dev.set_output(output=output_on)
            else:
                if use_meas_val:
                    volt_ac = self._dev.meas_volt_ac()
                    if volt_ac >= 0.2:
                        return False, 'Voltage too high to turn on/off.'
                else:
                    volt_set = self._dev.get_volt_ac()
                    if volt_set != 0:
                        return False, 'Voltage too high to turn on/off.'

                self._dev.set_output(output=output_on)

        return True, f'Set output for PCR: {params["output"]}'

    @ocs_agent.param('_')
    def get_output(self, session, params=None):
        """get_output()

        **Task** - Get output status.

        Notes
        -----
        The most recent data collected is stored in session.data in the
        structure::

            >>> response.session['data']
            {'output': True,
             'timestamp': 1737367649.2595236
            }
        """
        with self.lock.acquire_timeout(ACQ_TIMEOUT, job='get_output') as acquired:
            if not acquired:
                self.log.warn('Could not start get_output because '
                              f'{self.lock.job} is already running')
                return False, 'Could not acquire lock.'

            d_status = self._dev.get_output()
            session.data = {'output': d_status,
                            'timestamp': time.time()}

        return True, f'Got output status of PCR: {d_status}'

    @ocs_agent.param('volt_set', type=float, check=lambda x: 0 <= x <= VOLT_ULIM_SOFT)
    def set_volt_ac(self, session, params=None):
        """set_volt_ac(volt_set)

        **Task** - Set AC voltage value.

        Parameters
        ----------
        volt_set : float
            AC voltage setting in V. Values must be in range [0, 51].
        """
        with self.lock.acquire_timeout(ACQ_TIMEOUT, job='set_volt_ac') as acquired:
            if not acquired:
                self.log.warn('Could not start set_volt_ac because '
                              f'{self.lock.job} is already running')
                return False, 'Could not acquire lock.'

            self._dev.set_volt_ac(volt=params['volt_set'])

        return True, f'Set voltage for PCR: {params["volt_set"]}'

    @ocs_agent.param('_')
    def get_volt_ac(self, session, params=None):
        """get_volt_ac()

        **Task** - Get AC voltage setting.

        Notes
        -----
        The most recent data collected is stored in session.data in the
        structure::

            >>> response.session['data']
            {'volt_set': 10.0,
             'timestamp': 1737367649.2595236
            }
        """
        with self.lock.acquire_timeout(ACQ_TIMEOUT, job='get_volt_ac') as acquired:
            if not acquired:
                self.log.warn('Could not start get_volt_ac because '
                              f'{self.lock.job} is already running')
                return False, 'Could not acquire lock.'

            volt_set = self._dev.get_volt_ac()
            session.data = {'volt_set': volt_set,
                            'timestamp': time.time()}

        return True, f'Got AC voltage setting: {volt_set}'


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--port', default=PORT_DEFAULT, type=int,
                        help='Port number for TCP communication.')
    pgroup.add_argument('--ip-address',
                        help='IP address of the device.')

    return parser


def main(args=None):
    """Boot OCS agent"""
    txaio.start_logging(level=os.environ.get('LOGLEVEL', 'info'))

    parser = site_config.add_arguments()

    parser = make_parser()
    args = site_config.parse_args(agent_class='PCR500MAAgent',
                                  parser=parser,
                                  args=args)

    agent_inst, runner = ocs_agent.init_site_agent(args)

    pcr_agent = PCR500MAAgent(agent_inst, ip_addr=args.ip_address, port=args.port)

    agent_inst.register_task(
        'set_output',
        pcr_agent.set_output
    )

    agent_inst.register_task(
        'get_output',
        pcr_agent.get_output
    )

    agent_inst.register_task(
        'set_volt_ac',
        pcr_agent.set_volt_ac
    )

    agent_inst.register_task(
        'get_volt_ac',
        pcr_agent.get_volt_ac
    )

    agent_inst.register_process(
        'acq',
        pcr_agent.acq,
        pcr_agent._stop_acq,
        startup=True
    )

    runner.run(agent_inst, auto_reconnect=True)


if __name__ == '__main__':
    main()



"""
Interface to HWPSupervisor.

"""

import time
from dataclasses import asdict, dataclass, field

from ocs import client_http, ocs_client, site_config

# Helper functions for dealing with el ranges.


def _range_contains(a, b):
    # true if "a contains b" i.e. if b is a subset of a
    return a[0] <= b[0] and a[1] >= b[1]


def _simplify_ranges(*a):
    """Combine and order a set of intervals, to form an ordered,
    non-abutting list of intervals.

    """
    reduced = sorted([tuple(_a) for _a in a])
    i0 = 1
    while i0 < len(reduced):
        a, b = reduced[i0 - 1], reduced[i0]
        if a[1] >= b[0]:
            reduced[i0 - 1] = (a[0], max(b[1], a[1]))
            reduced.pop(i0)
        else:
            i0 += 1
    return reduced


AXIS_NAMES = ['el', 'az', 'third']


@dataclass
class MotionRule:
    #: Tuple giving the elevation range to which this rule applies.
    el_range: tuple

    #: List of grip_state values to which this rule applies. Note "*"
    #: matches any grip_state.
    grip_states: list

    #: List of spin_state values to which this rule applies. Note "*"
    #: matches any spin_state.
    spin_states: list

    #: Dict mapping each axis to a bool that indicates whether moves in
    #: that axis are permitted.
    allow_moves: dict

    def test_hwp(self, grip_state, spin_state):
        """Returns True only if grip_state and spin_state args are
        matched by this rule.

        """
        if grip_state not in self.grip_states and '*' not in self.grip_states:
            return False
        if spin_state not in self.spin_states and '*' not in self.spin_states:
            return False
        return True


@dataclass
class HWPInterlocks:
    #: bool indicating whether the config is valid (and thus eligible
    #: to be enabled).
    configured: bool = False

    #: bool indicating whether to bother with HWP interlocks.
    enabled: bool = False

    #: instance_id of HWPSupervisor agent to query.
    instance_id: str = 'hwp-supervisor'

    #: amount of leeway in el (deg) to grant when checking rules.
    tolerance: float = 0.1

    #: whether HWP state-based elevation limits should be applied to
    #: Sun avoidance escape movements
    limit_sun_avoidance: bool = True

    #: rules: list of MotionRule objects, each of which grants move
    #: privs on some axis based on elevation range and HWP state.
    rules: list = field(default_factory=list)

    @classmethod
    def from_dict(cls, cfg):
        if cfg is None:
            return cls()
        self = cls(configured=True, **cfg)
        self.rules = [MotionRule(**rule) for rule in self.rules]
        return self

    def encoded(self, basic=False):
        d = asdict(self)
        if basic:
            d.pop('rules')
        return d

    def get_client(self):
        return HWPSupervisorClient(self.instance_id)

    def test_range(self, el_range, grip_state, spin_state):
        """Determine what motion types are permitted, given the
        current HWP state and the elevation (range) of the planned
        motion.

        Args:
          el_range (tuple[float] or None):
            The starting and ending elevations of the move.
          grip_state (str):
            The HWP gripper state.
          spin_state (str):
            The HWP spinning state.

        The rules are evaluated in the context of grip_state and
        spin_state, to obtain the ranges of els over which moves in
        each axis is permitted.  Then the el_range, if provided is
        tested against those ranges.

        E.g for a call like::

          test_range((40, 40), 'ungripped', 'not_spinning')

        the returned structure looks like this::

          {'el': (True, (30, 70), [(30, 70)]),
           'az': (True, (30, 70), [(30, 70)]),
           'third': (False, None, [(30, 50), (60, 70)]),
          }

        I.e. for each axis, a tuple is returned where the third value
        is the full list of permitted elevation ranges for this state;
        the first value is whether the requested el_range,
        specifically, overlaps with the elevation ranges, and the
        second value is the particular range that overlaps with
        requested el_range (or None if not the case).

        """
        allow_moves = {k: [None, None, []] for k in AXIS_NAMES}
        for rule in self.rules:
            if not rule.test_hwp(grip_state, spin_state):
                continue
            # For each axis move that is allowed, add the el range
            # into the allow_moves entry.
            for k, (_, _, el_ranges) in allow_moves.items():
                if rule.allow_moves[k]:
                    el_ranges.append(rule.el_range)
        for k in allow_moves:
            allow_moves[k][2] = _simplify_ranges(*allow_moves[k][2])
        if el_range is not None:
            el_range = [min(el_range), max(el_range)]
            for k in allow_moves.keys():
                allow_moves[k][0] = False
                for _e in allow_moves[k][2]:
                    if _range_contains(_e, el_range):
                        allow_moves[k][0] = True
                        allow_moves[k][1] = _e
        return {k: tuple(v) for k, v in allow_moves.items()}


class HWPSupervisorClient:
    """OCSClient wrapper to query and interpret the state of the
    HWPSupervisor.  Instantiate with the instance_id of the
    HWPSupervisor to be monitored.

    """

    def __init__(self, instance_id):
        self.instance_id = instance_id
        self.cclient = site_config.get_control_client(instance_id)

    def update(self):
        """Analyze HWPSupervisor state info and make determinations
        about hwp state.  Returns a dict with the results.

        When everything works well, the useful results in the dict are:

        - ``timestamp``: current time
        - ``ok``: True if state determination went reasonably well
        - ``err_msg``: populated if not ok.
        - ``grip_state``: one of "unknown", "cold", "warm", "ungripped"
        - ``spin_state``: one of "unknown", "spinning", "not_spinning"

        The following raw data from monitor process are copied into
        the output dict::

          hwp_state:
            is_spinning       -> _is_spinning
            pid_target_freq   -> _target_freq
            gripper:
              grip_state      -> _grip_state
              brake           -> _grip_brakes
          acu:
            request_block_motion             -> request_block_motion
            request_block_motion_timestamp   -> request_block_motion_timestamp

        """
        ok, err_msg = False, ''
        hs, acu = {}, {}
        try:
            _, _, session = ocs_client.OCSReply(*self.cclient.request('status', 'monitor'))
            if session is not None:
                hs = session['data'].get('hwp_state', {})
                acu = session['data'].get('acu', {})
                ok = True
        except client_http.ControlClientError as e:
            err_msg = f'Error getting status: {e}'
        except Exception as e:
            err_msg = f'Surprising error: {e}'

        # Enhanced spin state logic...
        is_spinning = hs.get('is_spinning')  # bool or None
        target_freq = hs.get('pid_target_freq')  # float or None I guess
        if is_spinning or target_freq not in [None, 0]:
            spin_state = 'spinning'
        elif is_spinning is False and target_freq in [0.]:
            spin_state = 'not_spinning'
        else:
            spin_state = 'unknown'

        # Enhanced gripper state logic...
        _grip_state = hs.get('gripper', {}).get('grip_state')
        brakes = hs.get('gripper', {}).get('brake')  # 1 1 1 when stable
        brakes_on = False
        try:
            brakes_on = len(brakes) and all(brakes)
        except BaseException:
            pass
        if _grip_state is not None and brakes_on:
            grip_state = _grip_state
        else:
            grip_state = 'unknown'

        return {
            'timestamp': time.time(),
            'ok': ok,
            'err_msg': err_msg,
            'grip_state': grip_state,
            'spin_state': spin_state,
            '_grip_brakes': brakes,
            '_grip_state': _grip_state,
            '_is_spinning': is_spinning,
            '_target_freq': target_freq,
            'request_block_motion': acu.get('request_block_motion'),
            'request_block_motion_timestamp': acu.get('request_block_motion_timestamp'),
        }



import argparse
import os
import threading
import time
import traceback
from contextlib import contextmanager
from dataclasses import asdict, dataclass, field
from typing import Any, Dict, Generator, List, Literal, Optional, Tuple

import numpy as np
import ocs
import txaio
from ocs import client_http, ocs_agent, site_config
from ocs.client_http import ControlClient, ControlClientError
from ocs.ocs_client import OCSClient, OCSReply
from ocs.ocs_twisted import Pacemaker

MAX_SPIN_UP_DURATION: float = 1800.

client_cache: Dict[str, ControlClient] = {}


def get_op_data(agent_id, op_name, log=None, test_mode=False):
    """
    Process data from an agent operation, and formats it for the ``monitor``
    operation session data.

    Parameters
    --------------
    agent_id : str
        Instance ID of the agent
    op_name : str
        Operation from which to grab session data
    log : logger, optional
        Log object
    test_mode : bool
        If True, this will run in test mode, and not try to connect to the
        specified agent.

    Returns
    -----------
    Returns a dictionary with the following fields:

    agent_id: str
        Instance id for the agent being queried
    op_name : str
        Operation name being queried
    timestamp : float
        Time the operation status was queried
    data : dict
        Session data of the operation. This will be ``None`` if we can't connect
        to the operation.
    status : str
        Connection status of the operation. This can be the following:

        - ``no_agent_provided``: if the passed agent_id is None
        - ``test_mode``: if ``test_mode=True``, this will be the status and no
          operation will be queried.
        - ``op_not_found``: This means the operation could not be found, likely
          meaning the agent isn't running
        - ``no_active_session``: This means the operation specified exists but
           was never run.
        - ``ok``: Operation and session.data exist
    """
    if log is None:
        log = txaio.make_logger()  # pylint: disable=E1101

    data = {
        'agent_id': agent_id,
        'op_name': op_name,
        'timestamp': time.time(),
        'data': None
    }
    if agent_id is None:
        data['status'] = 'no_agent_provided'
        return data

    if test_mode:
        data['status'] = 'test_mode'
        return data

    args = []
    if 'SITE_HTTP' in os.environ:
        args += [f"--site-http={os.environ['SITE_HTTP']}"]

    if agent_id in client_cache:
        client = client_cache[agent_id]
    else:
        client = site_config.get_control_client(agent_id, args=args)
        client_cache[agent_id] = client

    try:
        _, _, session = OCSReply(*client.request('status', op_name))
    except client_http.ControlClientError as e:
        log.warn('Error getting status: {e}', e=e)
        data['status'] = 'op_not_found'
        return data

    if not session:
        data['status'] = 'no_active_session'
        return data

    data['data'] = session['data']
    data['status'] = 'ok'
    return data


@dataclass
class HWPClients:
    encoder: Optional[OCSClient] = None
    pmx: Optional[OCSClient] = None
    pid: Optional[OCSClient] = None
    ups: Optional[OCSClient] = None
    lakeshore: Optional[OCSClient] = None
    gripper_iboot: Optional[OCSClient] = None
    driver_iboot: Optional[OCSClient] = None
    pcu: Optional[OCSClient] = None
    gripper: Optional[OCSClient] = None


@dataclass
class GripperState:
    """
    Attributes
    -------------
    instance_id: str
        Instance ID of gripper agent being monitored
    limit_warm_grip_state: List[bool]
        State of the warm-grip limit switches for the three actuators
    limit_cold_grip_state: List[bool]
        State of the cold-grip limit switches for the three actuators
    emg: List[bool]
        For each actuator, this will be true if the actuator is receiving power
        from the motor controller.
    brake: List[bool]
        For each actuator, this will be true if the brake is enabled.
    grip_state: Literal['cold', 'warm', 'ungripped', 'unknown']
        A string representing the state of the gripper based on limit switches
        and other information.
    last_updated: Optional[float]
        Timestamp of last update.
    gripper_max_time_since_update: float
        Max amount of time without an update before the grip_state reverts to 'unknown'.
    """
    instance_id: str
    limit_warm_grip_state: List[bool] = field(default_factory=lambda: [False, False, False])
    limit_cold_grip_state: List[bool] = field(default_factory=lambda: [False, False, False])
    emg: List[bool] = field(default_factory=lambda: [False, False, False])
    brake: List[bool] = field(default_factory=lambda: [False, False, False])
    grip_state: Literal['cold', 'warm', 'ungripped', 'unknown'] = 'unknown'
    last_updated: Optional[float] = None
    gripper_max_time_since_update: float = 60.0

    def _verify_update_time(self) -> None:
        """
        Will check if gripper state has been updated within the allowed time.
        If not, this will set the grip_state to "unknown".
        """
        if self.last_updated is None:
            self.grip_state = 'unknown'
        elif time.time() - self.last_updated > self.gripper_max_time_since_update:
            self.grip_state = 'unknown'

    def update(self) -> None:
        op = get_op_data(self.instance_id, 'monitor_state', test_mode=False)
        if op['status'] != 'ok':
            self._verify_update_time()
            return

        d = op['data']
        state = d['state']
        self.last_updated = d['last_updated']

        for i, axis in enumerate([1, 2, 3]):
            self.limit_warm_grip_state[i] = state[f'act{axis}_limit_warm_grip_state']
            self.limit_cold_grip_state[i] = state[f'act{axis}_limit_cold_grip_state']
            self.emg[i] = state[f'act{axis}_emg']
            self.brake[i] = state[f'act{axis}_brake']

        if np.all(self.limit_cold_grip_state):
            self.grip_state = 'cold'
        elif np.any(self.limit_cold_grip_state):
            self.grip_state = 'unknown'
        elif np.all(self.limit_warm_grip_state):
            self.grip_state = 'warm'
        elif np.any(self.limit_warm_grip_state):
            self.grip_state = 'unknown'
        else:
            self.grip_state = 'ungripped'

        self._verify_update_time()


@dataclass
class IBootState:
    instance_id: str
    outlets: List[int]
    agent_type: Literal['iboot, synaccess']
    outlet_state: Dict[int, Optional[int]] = None
    op_data: Optional[Dict] = None

    def __post_init__(self):
        self.outlet_state = {o: None for o in self.outlets}

    def update(self):
        op = get_op_data(self.instance_id, 'acq', test_mode=False)
        self.op_data = op
        if op['status'] != 'ok':
            self.outlet_state = {o: None for o in self.outlets}
            return

        if self.agent_type == 'iboot':
            self.outlet_labels = {o: f'outletStatus_{o - 1}' for o in self.outlets}
            self.outlet_state = {
                outlet: op['data'][label]['status']
                for outlet, label in self.outlet_labels.items()
            }
        elif self.agent_type == 'synaccess':
            self.outlet_labels = {o: str(o - 1) for o in self.outlets}
            self.outlet_state = {
                outlet: op['data']['fields'][label]['status']
                for outlet, label in self.outlet_labels.items()
            }
        else:
            raise ValueError(
                f"Invalid agent_type: {self.agent_type}. "
                "Must be in ['iboot', 'synaccess']"
            )


@dataclass
class ACUState:
    """
    Class containing ACU state information.

    Args
    ------
    instance_id : str
        Instance ID of ACU agent
    min_el : float
        Minimum elevation allowed before restricting spin-up [deg]
    max_el : float
        Maximum elevation allowed before restricting spin-up [deg]
    max_time_since_update : float
        Maximum time since last update before restricting spin-up[sec]
    mount_velocity_grip_thresh: float
        Maximum mount velocity for gripping the HWP [deg/s] .
    grip_max_boresight_angle: float
        Maximum absolute boresight angle [deg] for which operating the grippers
        is allowed.

    Attributes
    ------------
    el_current_position : float
        Current el position [deg]
    el_commanded_position : float
        Commanded el position [deg]
    el_current_velocity : float
        Current el velocity [deg/s]
    az_current_velocity : float
        Current az velocity [deg/s]
    bor_current_position : float
        Current bor position [deg]
    bor_commanded_position : float
        Commanded bor position [deg]
    last_updated : float
        Time of last update [sec]. If this is None, you cannot trust other
        state variables.
    """
    instance_id: str
    min_el: float
    max_el: float
    max_time_since_update: float
    mount_velocity_grip_thresh: float
    grip_max_boresight_angle: float

    last_updated: Optional[float] = None
    el_current_position: float = 0.0
    el_commanded_position: float = 0.0
    el_current_velocity: float = 0.0
    az_current_velocity: float = 0.0
    bor_current_position: float = 0.0
    bor_commanded_position: float = 0.0

    # This will only be set by spin-control agent for communication with ACU
    request_block_motion: bool = False
    request_block_motion_timestamp: float = 0.0

    ACU_motion_blocked: Optional[bool] = None  # This flag is only set by the ACU agent
    use_acu_blocking: bool = False
    block_motion_timeout: float = 60.0

    def set_request_block_motion(self, state: bool) -> None:
        self.request_block_motion = state
        self.request_block_motion_timestamp = time.time()

    def update(self):
        op = get_op_data(self.instance_id, 'monitor')
        if op['status'] != 'ok':
            return

        d = op['data'].get("StatusDetailed")
        if d is None:
            return

        self.el_current_position = d['Elevation current position']
        self.el_commanded_position = d['Elevation commanded position']
        self.el_current_velocity = d['Elevation current velocity']
        self.az_current_velocity = d['Azimuth current velocity']
        self.bor_current_position = d['Boresight current position']
        self.bor_commanded_position = d['Boresight commanded position']

        t = d.get('timestamp_agent')
        if t is None:
            t = time.time()
        self.last_updated = t


@dataclass
class HWPState:
    last_updated: Optional[float] = None

    lakeshore_instance_id: Optional[str] = None
    ups_instance_id: Optional[str] = None
    pid_instance_id: Optional[str] = None
    pmx_instance_id: Optional[str] = None
    enc_instance_id: Optional[str] = None

    temp: Optional[float] = None
    temp_status: Optional[str] = None
    temp_thresh: Optional[float] = None
    temp_field: Optional[str] = None

    ups_output_source: Optional[str] = None
    ups_estimated_minutes_remaining: Optional[float] = None
    ups_estimated_charge_remaining: Optional[float] = None
    ups_battery_voltage: Optional[float] = None
    ups_battery_current: Optional[float] = None
    ups_minutes_remaining_thresh: Optional[float] = None
    ups_connected: Optional[bool] = None
    ups_last_connection_attempt: Optional[bool] = None

    pid_current_freq: Optional[float] = None
    pid_freq_tolerance: float = 0.1
    pid_target_freq: Optional[float] = None
    pid_direction: Optional[str] = None
    pid_last_updated: Optional[float] = None
    pid_max_time_since_update: float = 60.0

    pmx_current: Optional[float] = None
    pmx_voltage: Optional[float] = None
    pmx_source: Optional[str] = None
    pmx_last_updated: Optional[float] = None

    enc_freq: Optional[float] = None
    encoder_last_updated: Optional[float] = None
    last_quad: Optional[float] = None
    last_quad_time: Optional[float] = None

    gripper_iboot: Optional[IBootState] = None
    driver_iboot: Optional[IBootState] = None

    acu: Optional[ACUState] = None
    gripper: Optional[GripperState] = None

    is_spinning: Optional[bool] = None
    direction: Optional[str] = None
    forward_is_cw: Optional[bool] = True

    supervisor_control_state: Optional[Dict[str, Any]] = None

    def __post_init__(self) -> None:
        self.lock: threading.Semaphore = threading.Semaphore()

    @classmethod
    def from_args(cls, args: argparse.Namespace) -> "HWPState":
        log = txaio.make_logger()  # pylint: disable=E1101
        self = cls(
            lakeshore_instance_id=args.ybco_lakeshore_id,
            ups_instance_id=args.ups_id,
            pid_instance_id=args.hwp_pid_id,
            pmx_instance_id=args.hwp_pmx_id,
            enc_instance_id=args.hwp_encoder_id,
            temp_field=args.ybco_temp_field,
            temp_thresh=args.ybco_temp_thresh,
            ups_minutes_remaining_thresh=args.ups_minutes_remaining_thresh,
            pid_max_time_since_update=args.pid_max_time_since_update,
            forward_is_cw=args.forward_dir == 'cw',
        )

        if args.gripper_iboot_id is not None:
            self.gripper_iboot = IBootState(args.gripper_iboot_id, args.gripper_iboot_outlets,
                                            args.gripper_power_agent_type)
            log.info("Gripper Ibootbar id set: {id}", id=args.gripper_iboot_id)
        else:
            log.warn("Gripper Ibootbar id not set")

        if args.driver_iboot_id is not None:
            self.driver_iboot = IBootState(args.driver_iboot_id, args.driver_iboot_outlets,
                                           args.driver_power_agent_type)
            log.info("Driver Ibootbar id set: {id}", id=args.driver_iboot_id)
        else:
            log.warn("Driver Ibootbar id not set")

        if not args.no_acu:
            self.acu = ACUState(
                instance_id=args.acu_instance_id,
                min_el=args.acu_min_el - args.acu_tol_el,
                max_el=args.acu_max_el + args.acu_tol_el,
                max_time_since_update=args.acu_max_time_since_update,
                mount_velocity_grip_thresh=args.mount_velocity_grip_thresh,
                grip_max_boresight_angle=args.grip_max_boresight_angle,
                use_acu_blocking=args.use_acu_blocking,
                block_motion_timeout=args.acu_block_motion_timeout,
            )
            log.info("ACU state checking enabled: instance_id={id}",
                     id=self.acu.instance_id)
        else:
            log.warn("The no-acu option has been set. ACU state checking disabled.")

        if args.hwp_gripper_id is not None:
            self.gripper = GripperState(args.hwp_gripper_id)
        else:
            log.warn("HWP Gripper ID is not set. This will not be monitored in the HWP state.")

        return self

    def _update_from_keymap(self, op, keymap):
        if op['status'] != 'ok':
            for k in keymap:
                setattr(self, k, None)
            return

        for k, v in keymap.items():
            setattr(self, k, op['data'].get(v))

    def update_enc_state(self, test_mode=False):
        """
        Updates state values from the encoder acq operation results.

        Args
        -----
        op : dict
            Dict containing the operations (from get_op_data) from the encoder
            ``acq`` process
        """
        op = get_op_data(self.enc_instance_id, 'acq', test_mode=test_mode)
        self._update_from_keymap(op, {
            'enc_freq': 'approx_hwp_freq',
            'encoder_last_updated': 'encoder_last_updated',
            'last_quad': 'last_quad',
            'last_quad_time': 'last_quad_time',
        })
        return op

    def update_temp_state(self, test_mode=False):
        """
        Updates state values from the Lakeshore acq operation results.

        Args
        -----
        op : dict
            Dict containing the operations (from get_op_data) from the lakeshore
            ``acq`` process
        """
        op = get_op_data(self.lakeshore_instance_id, 'acq', test_mode=test_mode)
        if op['status'] != 'ok':
            self.temp = None
            self.temp_status = 'no_data'
            return op

        fields = op['data']['fields']
        if self.temp_field not in fields:
            self.temp = None
            self.temp_status = 'no_data'
            return op

        self.temp = fields[self.temp_field]['T']

        if self.temp_thresh is not None:
            if self.temp > self.temp_thresh:
                self.temp_status = 'over'
            else:
                self.temp_status = 'ok'
        else:
            self.temp_status = 'ok'
        return op

    def update_pmx_state(self, test_mode=False):
        """
        Updates state values from the pmx main operation results.

        Args
        -----
        op : dict
            Dict containing the operations (from get_op_data) from the pmx
            ``main`` process
        """
        op = get_op_data(self.pmx_instance_id, 'main', test_mode=test_mode)
        keymap = {'pmx_current': 'curr', 'pmx_voltage': 'volt',
                  'pmx_source': 'source', 'pmx_last_updated': 'last_updated'}
        self._update_from_keymap(op, keymap)
        return op

    def update_pid_state(self, test_mode=False):
        """
        Updates state values from the pid main operation results.

        Args
        -----
        op : dict
            Dict containing the operations (from get_op_data) from the pid
            ``main`` process
        """
        op = get_op_data(self.pid_instance_id, 'main', test_mode=test_mode)
        self._update_from_keymap(op, {
            'pid_current_freq': 'current_freq',
            'pid_target_freq': 'target_freq',
            'pid_direction': 'direction',
            'pid_last_updated': 'last_updated'
        })
        return op

    def update_ups_state(self, test_mode=False):
        """
        Updates state values from the UPS acq operation results.

        Args
        -----
        op : dict
            Dict containing the operations (from get_op_data) from the UPS
            ``acq`` process
        """
        op = get_op_data(self.ups_instance_id, 'acq', test_mode=test_mode)
        ups_keymap = {
            'ups_output_source': ('upsOutputSource', 'description'),
            'ups_estimated_minutes_remaining': ('upsEstimatedMinutesRemaining', 'status'),
            'ups_estimated_charge_remaining': ('upsEstimatedChargeRemaining', 'status'),
            'ups_battery_voltage': ('upsBatteryVoltage', 'status'),
            'ups_battery_current': ('upsBatteryCurrent', 'status'),
        }

        if op['status'] != 'ok':
            for k in ups_keymap:
                setattr(self, k, None)
            return

        # get oid
        data = op['data']
        for k in data:
            if k.startswith('upsOutputSource'):
                ups_oid = k.split('_')[1]
                break
        else:
            raise ValueError('Could not find upsOutputSource OID')

        for k, f in ups_keymap.items():
            setattr(self, k, data[f'{f[0]}_{ups_oid}'][f[1]])

        self.ups_last_connection_attempt = data['ups_connection']['last_attempt']
        self.ups_connected = data['ups_connection']['connected']

    def update_spin_state(self):
        """
        Updates hwp spin state values from the pid_state and control_state.
        """
        if self.pid_last_updated is None or self.pid_current_freq is None or \
                self.pid_direction is None:
            self.is_spinning = None
            self.direction = None
        elif time.time() - self.pid_last_updated > self.pid_max_time_since_update:
            self.is_spinning = None
            self.direction = None
        elif self.pid_current_freq > self.pid_freq_tolerance:
            self.is_spinning = True
        else:
            self.is_spinning = False
            self.direction = None
        if self.is_spinning and self.supervisor_control_state is not None:
            control_state = self.supervisor_control_state['state_type']
            # If hwp is braking, pid_direction is different from actual direction.
            # Therefore, we keep previous value and do not update direction.
            if control_state not in ('Brake', 'WaitForBrake'):
                if self.pid_direction == 0:
                    self.direction = 'cw' if self.forward_is_cw else 'ccw'
                elif self.pid_direction == 1:
                    self.direction = 'ccw' if self.forward_is_cw else 'cw'

    @property
    def pmx_action(self):
        """
        PMX action to take based on the current state of the HWP. This can be
        ``stop``, ``ok``, or ``no_data``.
        """
        # First check if either ups or temp are beyond a threshold
        if self.temp is not None and self.temp_thresh is not None:
            if self.temp > self.temp_thresh:
                return 'stop'

        min_remaining = self.ups_estimated_minutes_remaining
        min_remaining_thresh = self.ups_minutes_remaining_thresh
        if min_remaining is not None and min_remaining_thresh is not None and \
                self.ups_output_source is not None:
            if min_remaining < min_remaining_thresh and \
                    self.ups_output_source != 'normal':
                return 'stop'

        # If either ybco_temp or ups state is None, return no_data
        if self.temp is None and (self.temp_thresh is not None):
            return 'no_data'

        if min_remaining is None and (min_remaining_thresh is not None):
            return 'no_data'

        return 'ok'

    def update(self, test_mode=False) -> Dict[str, Any]:
        now = time.time()
        with self.lock:
            ops = {
                'temperature': self.update_temp_state(test_mode=test_mode),
                'encoder': self.update_enc_state(test_mode=test_mode),
                'pmx': self.update_pmx_state(test_mode=test_mode),
                'pid': self.update_pid_state(test_mode=test_mode),
                'ups': self.update_ups_state(),
            }
            self.update_spin_state()

            self.last_updated = now
            if self.driver_iboot is not None:
                self.driver_iboot.update()
            if self.gripper_iboot is not None:
                self.gripper_iboot.update()
            if self.acu is not None:
                self.acu.update()
            if self.gripper is not None:
                self.gripper.update()

        return ops

    @property
    def gripper_action(self):
        """
        Gripper action to take based on the current state of the HWP. This can be
        ``stop``, ``ok``, or ``no_data``.
        """
        pmx_action = self.pmx_action
        if pmx_action == 'ok':
            return 'ok'
        if pmx_action == 'no_data':
            return 'no_data'

        if self.pid_current_freq is None:
            return 'no_data'
        elif self.pid_current_freq > 0.01:
            return 'ok'
        else:  # Only grip if the hwp_freq is smaller than 0.01
            return 'stop'


class ControlStateInfo:
    def __init__(self, state):
        """
        Class that holds the control state dataclass, and relevant metadata
        such as start time and last_update time.

        Args
        ------
        state : ControlState
            ControlState object
        """
        self.state = state
        self.last_update_time = 0
        self.start_time = time.time()
        self.state_type = state.__class__.__name__

    def encode(self):
        d = {
            'state_type': self.state_type,
            'start_time': self.start_time,
            'last_update_time': self.last_update_time,
        }
        d.update(asdict(self.state))
        return d


class ControlState:
    """Namespace for HWP control state definitions"""
    @dataclass
    class Idle:
        """Does nothing"""

    @dataclass
    class PIDToFreq:
        """
        Configures PID and PMX agents to PID to a target frequency.

        Attributes
        -----------
        target_freq : float
            Target frequency to PID to
        direction : str
            Direction to set the PID. Should either be '1' or '0'.
        freq_tol : float
            Tolerance between the target frequency and the current frequency
            to consider the target frequency reached.
        freq_tol_duration : float
            Duration in seconds that the frequency must be within the tolerance
        """
        target_freq: float
        direction: str
        freq_tol: float
        freq_tol_duration: float

    @dataclass
    class CheckInitialRotation:
        """
        In this state, will check if the HWP has started rotating. If it has not
        started rotating in ``check_wait_time`` seconds, it will briefly turn on the PCU
        to initiate rotation, before transitioning to the WaitForTargetFreq state.
        """
        target_freq: float
        freq_tol: float
        freq_tol_duration: float
        direction: str
        check_wait_time: float = 15.0
        start_time: float = field(default_factory=time.time)

    @dataclass
    class WaitForTargetFreq:
        """
        Wait until HWP reaches its target frequency before transitioning to
        the Done state.

        Attributes
        -----------
        target_freq : float
            Target frequency to PID to
        freq_tol : float
            Tolerance between the target frequency and the current frequency
            to consider the target frequency reached.
        freq_tol_duration : float
            Duration in seconds that the frequency must be within the tolerance
        freq_within_thresh_start : float
            Time that the frequency entered the tolerance range
        max_duration : float
            Maximum duration of time to wait in seconds
        start_time : float
            Time that the state was entered
        """
        target_freq: float
        freq_tol: float
        freq_tol_duration: float
        freq_within_tol_start: Optional[float] = None
        max_duration: Optional[float] = None
        start_time: float = field(default_factory=time.time)
        direction: str = ''
        _pcu_enabled: bool = field(init=False, default=False)

    @dataclass
    class ConstVolt:
        """
        Configure PMX agent to output a constant voltage.

        Attributes
        -----------
        voltage : float
            Voltage to set the PMX to
        start_time : float
            Time that the state was entered
        """
        voltage: float
        direction: str

    @dataclass
    class Done:
        """
        Signals the last state has completed

        Attributes
        -----------
        success : bool
            Whether the last state was completed successfully
        msg : str
            Optional message to include with the Done state
        start_time : float
            Time that the state was entered
        """
        success: bool
        msg: str = None

    @dataclass
    class Error:
        """
        Signals the last state update threw an error

        Attributes
        -----------
        traceback : str
            Traceback of the error
        start_time : float
            Time that the state was entered
        """
        traceback: str
        start_time: float = field(default_factory=time.time)

    @dataclass
    class Brake:
        """
        Configure the PID and PMX agents to actively brake the HWP
        """
        freq_tol: float
        freq_tol_duration: float
        brake_voltage: float

    @dataclass
    class WaitForBrake:
        """
        Waits until the HWP has slowed before shutting off PMX

        min_freq : float
            Frequency (Hz) below which the PMX should be shut off.
        """
        min_freq: float
        prev_freq: float = None

    @dataclass
    class PmxOff:
        """
        Turns off the PMX

        Attributes
        -----------
        wait_stop : bool
            Whether to wait until hwp stops
        freq_tol : float
            Tolerance of frequency to consider hwp is stopped.
        freq_tol_duration : float
            Duration in seconds that the frequency must be within the tolerance
        success : bool
            Whether the last state was completed successfully
        """
        wait_stop: bool = False
        freq_tol: float = 0.05
        freq_tol_duration: float = 30
        success: bool = True

    @dataclass
    class GripHWP:
        """
        Grips the HWP
        """

    @dataclass
    class UngripHWP:
        """
        Ungrips the HWP
        """

    @dataclass
    class Abort:
        """Abort current action"""
        pass

    @dataclass
    class EnableDriverBoard:
        """
        Enables driver boards for encoder LEDs

        Attributes
        -----------
        driver_power_agent_type : Literal["iboot", "synaccess"]
            Type of agent used for controlling the driver power. Must be
            "iboot" or "synaccess".
        outlets : List[int]
            Outlets required for enabling driver board
        cycle_twice : bool
            If true, will wait, and then power cycle again
        cycle_wait_time : float
            Time [sec] before repeating the power cycle
        """
        driver_power_agent_type: Literal['iboot, synaccess']
        outlets: List[int]
        cycle_twice: bool = False
        cycle_wait_time: float = 60 * 5

        def __post_init__(self):
            if self.driver_power_agent_type not in ['iboot', 'synaccess']:
                raise ValueError(
                    f"Invalid driver_power_agent_type: {self.driver_power_agent_type}. "
                    "Must be in ['iboot', 'synaccess']"
                )
            self.cycled = False
            self.cycle_timestamp = None

    @dataclass
    class DisableDriverBoard:
        """
        Disables driver board for encoder LEDs

        Attributes
        -----------
        driver_power_agent_type: Literal['iboot, synaccess']
            Type of agent used for controlling the driver power
        outlets: List[int]
            Outlets required for enabling driver board
        """
        driver_power_agent_type: Literal['iboot, synaccess']
        outlets: List[int]

        def __post_init__(self):
            if self.driver_power_agent_type not in ['iboot', 'synaccess']:
                raise ValueError(
                    f"Invalid driver_power_agent_type: {self.driver_power_agent_type}. "
                    "Must be in ['iboot', 'synaccess']"
                )

    completed_states = (Done, Error, Abort, Idle)


class ControlAction:
    """
    This is a class to contain data regarding a single HWP control action.
    This groups together states that are part of a single action, and whether
    the action is completed and successful.
    """
    _cur_action_id: int = 0
    _id_lock = threading.Lock()

    def __init__(self, state):
        with ControlAction._id_lock:
            self.action_id = ControlAction._cur_action_id
            ControlAction._cur_action_id += 1

        self.completed = False
        self.success = False
        self.state_history = []
        self.log = txaio.make_logger()  # pylint: disable=E1101
        self.set_state(state)

    def set_state(self, state):
        """
        Sets state for the current action. If this is a `completed_state`,
        will mark as complete.
        """
        self.cur_state_info = ControlStateInfo(state)
        self.state_history.append(self.cur_state_info)
        self.log.info(f"Setting state: {state}")
        if isinstance(state, ControlState.completed_states):
            self.completed = True
        if isinstance(state, ControlState.Done):
            self.success = state.success

    def encode(self):
        """Encodes this as a dict"""
        return dict(
            action_id=self.action_id,
            completed=self.completed,
            success=self.success,
            cur_state=self.cur_state_info.encode(),
            state_history=[s.encode() for s in self.state_history],
        )

    def sleep_until_complete(self, session=None, dt=1):
        """
        Sleeps until the action is complete.

        Args
        -----
        session: OpSession, optional
            If specified, this will set `session.data['action']` to the encoded
            action on each iteration to keep session data up to date
        dt: float, optional
            Time to sleep between iterations.
        """
        while True:
            if session is not None:
                session.data.update({'action': self.encode()})
            if self.completed:
                return
            time.sleep(dt)


@contextmanager
def ensure_grip_safety(hwp_state: HWPState, log: txaio.ILogger) -> Generator[None, None, None]:
    """
    Run required checks for gripper safety. This will check ACU parameters such
    as az/el position and velocity. If `use_acu_blocking` is set, this will
    also attempt to block out ACU motion for the duration of the operation.

    Args
    ------
    hwp_state : HWPState
        HWP state object. In addition to reading ACU state vars, this will
        set the `request_block_ACU_motion` flag if `use_acu_blocking` is set.
    timeout: float
        Timeout for waiting for the ACU blockout before an error will be raised.
    """
    now = time.time()

    if hwp_state.pid_current_freq is None or hwp_state.pid_last_updated is None:
        raise RuntimeError("Cannot determine current HWP Freq")

    tdiff = now - hwp_state.pid_last_updated
    if tdiff > hwp_state.pid_max_time_since_update:
        raise RuntimeError(f"HWP PID state has not been updated in {tdiff} sec")

    if np.abs(hwp_state.pid_current_freq) > 0.02:
        raise RuntimeError("Cannot grip HWP while spinning")
    log.info("Rotation safety checks have passed")

    acu = hwp_state.acu
    if acu is None:
        yield
        return

    # Run checks of ACU state
    if acu.last_updated is None:
        raise RuntimeError(
            f"No ACU data has been received from instance-id {acu.instance_id}"
        )

    tdiff = time.time() - acu.last_updated
    if tdiff > acu.max_time_since_update:
        raise RuntimeError(f"ACU state has not been updated in {tdiff} sec")

    if not (acu.min_el <= acu.el_current_position <= acu.max_el):
        raise RuntimeError(
            f"ACU elevation is {acu.el_current_position} deg, "
            f"outside of allowed range ({acu.min_el}, {acu.max_el})"
        )
    if not (acu.min_el <= acu.el_commanded_position <= acu.max_el):
        raise RuntimeError(
            f"ACU commanded elevation is {acu.el_commanded_position} deg, "
            f"outside of allowed range ({acu.min_el}, {acu.max_el})"
        )
    if np.abs(acu.az_current_velocity) > np.abs(acu.mount_velocity_grip_thresh):
        raise RuntimeError(
            f"ACU az-velocity is {acu.az_current_velocity} deg/s, "
            f"above threshold of {acu.mount_velocity_grip_thresh} deg/s"
        )
    if np.abs(acu.el_current_velocity) > np.abs(acu.mount_velocity_grip_thresh):
        raise RuntimeError(
            f"ACU el-velocity is {acu.el_current_velocity} deg/s, "
            f"above threshold of {acu.mount_velocity_grip_thresh} deg/s"
        )
    if np.abs(acu.bor_current_position) > np.abs(acu.grip_max_boresight_angle):
        raise RuntimeError(
            f"boresight current angle of {acu.bor_current_position} deg, "
            f"above threshold of {acu.grip_max_boresight_angle} deg"
        )
    if np.abs(acu.bor_commanded_position) > np.abs(acu.grip_max_boresight_angle):
        raise RuntimeError(
            f"boresight commanded angle of {acu.bor_commanded_position} deg, "
            f"above threshold of {acu.grip_max_boresight_angle} deg"
        )
    log.info("ACU safety checks have passed")

    if not acu.use_acu_blocking:
        yield
        return

    try:  # If use_acu_blocking, do handshake with ACU agent to block motino
        acu.set_request_block_motion(True)
        log.info("Requesting ACU to block motion")
        while not acu.ACU_motion_blocked:
            if time.time() - acu.request_block_motion_timestamp > acu.block_motion_timeout:
                raise RuntimeError("ACU motion was not blocked within timeout")
            time.sleep(1)
        log.info("ACU motion has been blocked")
        yield
    finally:
        log.info("Releasing ACU motion block")
        acu.set_request_block_motion(False)


class ControlStateMachine:
    def __init__(self) -> None:
        self.action: ControlAction = ControlAction(ControlState.Idle())
        self.action_history: List[ControlAction] = []
        self.max_action_history_count = 100
        self.log = txaio.make_logger()  # pylint: disable=E1101
        self.lock = threading.Lock()

    def run_and_validate(self, op, kwargs=None, timeout=30, log=None):
        """
        Runs an OCS Operation, and validates that it was successful.

        Args
        -------
        op : OCS MatchedOp
            Operation to run. This must be a MatchedOp, or a method of an
            OCSClient
        kwargs : dict, optional
            Kwargs to pass to the operation
        timeout : float, optional
            Timeout for the wait command. This defaults to
            ``default_wait_timeout`` which is 10 seconds. If this is set to
            None, will wait indefinitely.
        """
        if kwargs is None:
            kwargs = {}
        if log is None:
            log = self.log

        status, msg, session = op.start(**kwargs)
        log.info("Starting op: name={name}, kwargs={kw}",
                 name=session.get('op_name'), kw=kwargs)

        if status == ocs.ERROR:
            raise ControlClientError("op-start returned Error:\n  msg: " + msg)

        if status == ocs.TIMEOUT:
            raise ControlClientError("op-start timed out")

        status, msg, session = op.wait(timeout=timeout)

        if status == ocs.ERROR:
            raise ControlClientError("op-wait returned Error:\n  msg: " + msg)

        if status == ocs.TIMEOUT:
            raise ControlClientError("op-wait timed out")

        log.info("Completed op: name={name}, success={success}, kwargs={kw}",
                 name=session.get('op_name'), success=session.get('success'),
                 kw=kwargs)

        return session

    def update(self, clients, hwp_state: HWPState):
        """Run the next series of actions for the current state"""
        try:
            self.lock.acquire()
            state = self.action.cur_state_info.state
            self.action.cur_state_info.last_update_time = time.time()

            def query_pid_state():
                data = self.run_and_validate(clients.pid.get_state)['data']
                self.log.info("pid state: {data}", data=data)
                return data

            def check_acu_ok_for_spinup():
                acu = hwp_state.acu
                if acu is not None:
                    if acu.last_updated is None:
                        raise RuntimeError(f"No ACU data has been received from instance-id {acu.instance_id}")
                    tdiff = time.time() - acu.last_updated
                    if tdiff > acu.max_time_since_update:
                        raise RuntimeError(f"ACU state has not been updated in {tdiff} sec")
                    if not (acu.min_el <= acu.el_current_position <= acu.max_el):
                        raise RuntimeError(f"ACU elevation is {acu.el_current_position} deg, "
                                           f"outside of allowed range ({acu.min_el}, {acu.max_el})")
                    if not (acu.min_el <= acu.el_commanded_position <= acu.max_el):
                        raise RuntimeError(f"ACU commanded elevation is {acu.el_commanded_position} deg, "
                                           f"outside of allowed range ({acu.min_el}, {acu.max_el})")

            def check_gripper_ok_for_spinup():
                if hwp_state.gripper is None:  # No gripper was specified in the config file -- ignore
                    return
                grip_state = hwp_state.gripper.grip_state
                if grip_state != "ungripped":
                    raise RuntimeError(
                        f"Spinup attempted while grip state is: {grip_state}"
                    )

            if isinstance(state, ControlState.PIDToFreq):
                check_acu_ok_for_spinup()
                check_gripper_ok_for_spinup()
                self.run_and_validate(clients.pid.set_direction,
                                      kwargs={'direction': state.direction})
                self.run_and_validate(clients.pid.declare_freq,
                                      kwargs={'freq': state.target_freq})
                self.run_and_validate(clients.pmx.use_ext)
                self.run_and_validate(clients.pmx.set_on)
                self.run_and_validate(clients.pid.tune_freq, timeout=60)
                self.run_and_validate(
                    clients.pcu.send_command,
                    kwargs={'command': 'off'}, timeout=None,
                )
                self.action.set_state(ControlState.CheckInitialRotation(
                    target_freq=state.target_freq,
                    freq_tol=state.freq_tol,
                    freq_tol_duration=state.freq_tol_duration,
                    direction=state.direction,
                ))

            if isinstance(state, ControlState.CheckInitialRotation):
                fhwp = hwp_state.enc_freq
                if fhwp >= 0.2:
                    self.action.set_state(ControlState.WaitForTargetFreq(
                        target_freq=state.target_freq,
                        freq_tol=state.freq_tol,
                        freq_tol_duration=state.freq_tol_duration,
                        max_duration=MAX_SPIN_UP_DURATION,
                        direction=state.direction,
                    ))
                    return

                if time.time() - state.start_time < state.check_wait_time:
                    return

                if int(state.direction) == 1:  # Reverse
                    self.run_and_validate(
                        clients.pcu.send_command,
                        kwargs={'command': 'on_1'}, timeout=None
                    )
                else:
                    self.run_and_validate(
                        clients.pcu.send_command,
                        kwargs={'command': 'on_2'}, timeout=None
                    )
                time.sleep(3)
                self.run_and_validate(
                    clients.pcu.send_command,
                    kwargs={'command': 'off'}, timeout=None
                )
                self.action.set_state(ControlState.WaitForTargetFreq(
                    target_freq=state.target_freq,
                    freq_tol=state.freq_tol,
                    freq_tol_duration=state.freq_tol_duration,
                    max_duration=MAX_SPIN_UP_DURATION,
                    direction=state.direction,
                ))

            elif isinstance(state, ControlState.WaitForTargetFreq):
                # Check if we are close enough to the target frequency.
                # This will make sure we remain within the frequency threshold for
                # ``self.freq_tol_duration`` seconds before switching to DONE
                f = hwp_state.pid_current_freq

                # Enable pcu if spinning up faster than 1.5 Hz
                if state.target_freq > 1.5 and f > 1.0 and not state._pcu_enabled:
                    self.log.info("Enabling PCU")
                    if int(state.direction) == 1:  # Reverse
                        self.run_and_validate(
                            clients.pcu.send_command,
                            kwargs={'command': 'on_1'}, timeout=None
                        )
                    else:
                        self.run_and_validate(
                            clients.pcu.send_command,
                            kwargs={'command': 'on_2'}, timeout=None
                        )
                    state._pcu_enabled = True

                # If the frequency doen't get close enough within max diration
                # power off
                if state.max_duration is not None:
                    if time.time() - state.start_time > state.max_duration:
                        self.action.set_state(ControlState.PmxOff())

                if f is None:
                    state.freq_within_tol_start = None
                    return

                if np.abs(f - state.target_freq) > state.freq_tol:
                    state.freq_within_tol_start = None
                    return

                # If within tolerance for freq_tol_duration, switch to Done
                if state.freq_within_tol_start is None:
                    state.freq_within_tol_start = time.time()

                time_within_tol = time.time() - state.freq_within_tol_start
                if time_within_tol > state.freq_tol_duration:
                    self.action.set_state(ControlState.Done(success=True))

            elif isinstance(state, ControlState.ConstVolt):
                if state.voltage > 0:
                    check_acu_ok_for_spinup()
                    check_gripper_ok_for_spinup()
                self.run_and_validate(clients.pmx.set_on)
                self.run_and_validate(clients.pid.set_direction,
                                      kwargs={'direction': state.direction})
                self.run_and_validate(clients.pmx.ign_ext)
                self.run_and_validate(clients.pmx.set_v,
                                      kwargs={'volt': state.voltage})
                self.action.set_state(ControlState.Done(success=True))

            elif isinstance(state, ControlState.PmxOff):
                self.run_and_validate(clients.pmx.set_off)
                self.run_and_validate(clients.pid.declare_freq,
                                      kwargs={'freq': 0})
                self.run_and_validate(clients.pid.tune_freq)
                self.run_and_validate(
                    clients.pcu.send_command,
                    kwargs={'command': 'stop'}, timeout=None
                )
                if state.wait_stop:
                    self.action.set_state(ControlState.WaitForTargetFreq(
                        target_freq=0,
                        freq_tol=state.freq_tol,
                        freq_tol_duration=state.freq_tol_duration,
                    ))
                else:
                    self.action.set_state(ControlState.Done(success=state.success))

            elif isinstance(state, ControlState.GripHWP):
                with ensure_grip_safety(hwp_state, self.log):
                    # Grip can take 5-10 minutes to complete
                    self.run_and_validate(clients.gripper.grip, timeout=720)
                self.action.set_state(ControlState.Done(success=True))

            elif isinstance(state, ControlState.UngripHWP):
                with ensure_grip_safety(hwp_state, self.log):
                    self.run_and_validate(clients.gripper.ungrip, timeout=60)
                self.action.set_state(ControlState.Done(success=True))

            elif isinstance(state, ControlState.Brake):
                self.run_and_validate(
                    clients.pcu.send_command,
                    kwargs={'command': 'off'}, timeout=None
                )

                # Flip PID direciton and tune stop
                pid_dir = int(query_pid_state()['direction'])
                new_d = '0' if (pid_dir == 1) else '1'
                self.run_and_validate(clients.pid.set_direction,
                                      kwargs=dict(direction=new_d))
                self.run_and_validate(clients.pid.tune_stop)

                self.run_and_validate(clients.pmx.ign_ext)
                self.run_and_validate(clients.pmx.set_v, kwargs={'volt': state.brake_voltage})
                self.run_and_validate(clients.pmx.set_on)

                time.sleep(10)
                self.action.set_state(ControlState.WaitForBrake(
                    min_freq=0.5,
                    prev_freq=hwp_state.enc_freq,
                ))

            elif isinstance(state, ControlState.WaitForBrake):
                f0 = query_pid_state()['current_freq']
                time.sleep(5)
                f1 = query_pid_state()['current_freq']
                if f0 < 0.5 or (f1 > f0):
                    self.log.info("Turning off PMX and putting PCU in stop mode")
                    self.run_and_validate(clients.pmx.set_off)
                    self.run_and_validate(
                        clients.pcu.send_command,
                        kwargs={'command': 'stop'}, timeout=None
                    )
                    self.action.set_state(ControlState.WaitForTargetFreq(
                        target_freq=0,
                        freq_tol=0.05,
                        freq_tol_duration=30,
                    ))
                    return

            elif isinstance(state, ControlState.EnableDriverBoard):
                def set_outlet_state(outlet: int, outlet_state: bool):
                    if state.driver_power_agent_type == 'iboot':
                        kw = {'outlet': outlet, 'state': 'on' if outlet_state else 'off'}
                    else:
                        kw = {'outlet': outlet, 'on': outlet_state}
                    self.run_and_validate(clients.driver_iboot.set_outlet, kwargs=kw)

                if not state.cycled:
                    for outlet in state.outlets:
                        set_outlet_state(outlet, True)
                    state.cycled = True
                    state.cycle_timestamp = time.time()
                    if not state.cycle_twice:
                        self.action.set_state(ControlState.Done(success=True))
                        return

                # Needs to be re-cycled, after wait time
                if time.time() - state.cycle_timestamp < state.cycle_wait_time:
                    return

                for outlet in state.outlets:
                    set_outlet_state(outlet, False)
                time.sleep(5)
                for outlet in state.outlets:
                    set_outlet_state(outlet, True)
                self.action.set_state(ControlState.Done(success=True))
                return

            elif isinstance(state, ControlState.DisableDriverBoard):
                def set_outlet_state(outlet: int, outlet_state: bool):
                    if state.driver_power_agent_type == 'iboot':
                        kw = {'outlet': outlet, 'state': 'on' if outlet_state else 'off'}
                    else:
                        kw = {'outlet': outlet, 'on': outlet_state}
                    self.run_and_validate(clients.driver_iboot.set_outlet, kwargs=kw)

                for outlet in state.outlets:
                    set_outlet_state(outlet, False)
                self.action.set_state(ControlState.Done(success=True))
                return

        except Exception:
            tb = traceback.format_exc()
            self.log.error("Error updating state:\n{tb}", tb=tb)
            self.action.set_state(ControlState.Error(traceback=tb))
        finally:
            self.lock.release()

    def request_new_action(self, state):
        """
        Requests that a new action is started with a given state.
        If an action is already in progress, it will be aborted.
        """
        with self.lock:
            if not self.action.completed:
                self.action.set_state(ControlState.Abort())
            if len(self.action_history) > self.max_action_history_count:
                self.action_history.pop(0)
            self.action = ControlAction(state)
            self.action_history.append(self.action)
            return self.action


class HWPSupervisor:
    """
    The HWPSupervisor agent is responsible for monitoring HWP and related
    components, and high-level control of the HWP. This maintains an updated
    HWPState containing state info pertaining to HWP agents. Additionally, it
    contains a state-machine for controlling the HWP based on the ControlState.


    Attributes
    ----------------
    agent : OCSAgent
        OCS agent instance
    args : argparse.Namespace
        Argument namespace
    hwp_state : HWPState
        Class containing most recent state information for the HWP
    control_state : ControlState
        Current control_state object. This will be used to determine what
        commands to issue to HWP agents.
    forward_is_cw : bool
        True if the PID "forward" direction is clockwise, False if CCW.
    """

    def __init__(self, agent, args):
        self.agent = agent
        self.args = args

        self.sleep_time = args.sleep_time
        self.log = agent.log

        self.ybco_lakeshore_id = args.ybco_lakeshore_id
        self.ybco_temp_field = args.ybco_temp_field
        self.ybco_temp_thresh = args.ybco_temp_thresh

        self.hwp_encoder_id = args.hwp_encoder_id
        self.hwp_pmx_id = args.hwp_pmx_id
        self.hwp_pid_id = args.hwp_pid_id
        self.ups_id = args.ups_id

        self.hwp_state = HWPState.from_args(args)
        self.gripper_iboot_id = args.gripper_iboot_id
        self.driver_iboot_id = args.driver_iboot_id
        self.control_state_machine = ControlStateMachine()
        self.forward_is_cw = args.forward_dir == 'cw'

        self.driver_power_agent_type = args.driver_power_agent_type
        self.driver_iboot_outlets = args.driver_iboot_outlets
        self.driver_power_cycle_twice = args.driver_power_cycle_twice
        self.driver_power_cycle_wait_time = args.driver_power_cycle_wait_time

    def _get_hwp_clients(self):
        def get_client(id):
            args = []
            if 'SITE_HTTP' in os.environ:
                args += [f"--site-http={os.environ['SITE_HTTP']}"]
            if id is None:
                return None
            try:
                return OCSClient(id, args=args)
            except ControlClientError:
                self.log.error("Could not connect to client: {id}", id=id)
                return None

        return HWPClients(
            encoder=get_client(self.hwp_encoder_id),
            pmx=get_client(self.hwp_pmx_id),
            pid=get_client(self.hwp_pid_id),
            pcu=get_client(self.args.hwp_pcu_id),
            ups=get_client(self.ups_id),
            lakeshore=get_client(self.ybco_lakeshore_id),
            gripper_iboot=get_client(self.gripper_iboot_id),
            driver_iboot=get_client(self.driver_iboot_id),
            gripper=get_client(self.args.hwp_gripper_id),
        )

    def query_hwp_state(self, session, params) -> Tuple[bool, str]:
        """query_hwp_state()

        **Task** -- Forces an update of the HWP State, and returns the result.

        For information on session.data["state"], see the docstrings for the
        ``monitor`` process.
        """
        self.hwp_state.update()
        session.data['state'] = asdict(self.hwp_state)
        return True, "Queried HWP state"

    @ocs_agent.param('test_mode', type=bool, default=False)
    def monitor(self, session, params) -> Tuple[bool, str]:
        """monitor()

        **Process** -- Monitors various HWP related HK systems.

        This operation has three main steps:

        - Query session data for all HWP and HWP adjacent agents. Session info for each
          queried operation will be stored in the ``monitored_sessions`` field of the
          session data. See the docs for the ``get_op_data`` function for information on
          what info will be saved.
        - Parse session-data from monitored operations to create the ``state`` dict,
          containing info such as ``ybco_temp`` and ``hwp_freq``.
        - Determine subsystem actions based on the HWP state, which will be stored in
          the ``actions`` dict which can be read by hwp subsystems to initiate a
          shutdown

        An example of the session data, along with possible options for status
        strings can be seen below::

            >>> response.session['data']

                {'timestamp': 1601924482.722671,
                'monitored_sessions': {
                    'encoder': {
                        'agent_id': 'test',
                        'data': <session data for test.acq>,
                        'op_name': 'acq',
                        'status': 'ok',  # See ``get_op_data`` docstring for choices
                        'timestamp': 1680273288.6200094},
                    },
                    'temperature': {see above},
                    'ups': {see above},
                    'pmx': {see above},
                    'pid': {see above}}
                # State data parsed from monitored sessions
                'state': {
                    'acu': None,
                    'driver_iboot': None,
                    'enc_freq': None,
                    'enc_instance_id': 'hwp-enc',
                    'encoder_last_updated': None,
                    'gripper': {
                        'brake': [0, 0, 0],
                        'emg': [0, 0, 0],
                        'grip_state': 'ungripped',
                        'gripper_max_time_since_update': 60.0,
                        'instance_id': 'hwp-gripper',
                        'last_updated': 1737132296.0509753,
                        'limit_cold_grip_state': [0, 0, 0],
                        'limit_warm_grip_state': [0, 0, 0]},
                    'gripper_iboot': None,
                    'is_spinning': False,
                    'lakeshore_instance_id': None,
                    'last_quad': None,
                    'last_quad_time': None,
                    'last_updated': 1737132297.6445067,
                    'pid_current_freq': 0.0,
                    'pid_direction': 0,
                    'pid_freq_tolerance': 0.1,
                    'pid_instance_id': 'hwp-pid',
                    'pid_last_updated': 1737132293.4787645,
                    'pid_max_time_since_update': 60.0,
                    'pid_target_freq': 0.0,
                    'pmx_current': 0.0,
                    'pmx_instance_id': 'hwp-pmx',
                    'pmx_last_updated': 1737132292.9470851,
                    'pmx_source': 'volt',
                    'pmx_voltage': 0.0,
                    'supervisor_control_state': {
                        'last_update_time': 1737132297.3006465,
                        'start_time': 1737132296.9819815,
                        'state_type': 'Idle'},
                    'temp': None,
                    'temp_field': None,
                    'temp_status': 'no_data',
                    'temp_thresh': None,
                    'ups_battery_current': 0,
                    'ups_battery_voltage': 136,
                    'ups_connected': None,
                    'ups_estimated_charge_remaining': 50,
                    'ups_estimated_minutes_remaining': 45,
                    'ups_instance_id': ups,
                    'ups_last_connection_attempt': None,
                    'ups_minutes_remaining_thresh': None,
                    'ups_output_source': None}

                 # Subsystem action recommendations determined from state data
                'actions': {
                    'pmx': 'ok'  # 'ok', 'stop', or 'no_data'
                    'gripper': 'ok'  # 'ok', 'stop', or 'no_data'
                }}
        """
        pm = Pacemaker(1. / self.sleep_time)
        test_mode = params.get('test_mode', False)

        session.data = {
            'timestamp': time.time(),
            'monitored_sessions': {},
            'hwp_state': {},
            'actions': {
                'pmx': 'no_data',
                'gripper': 'no_data',
            }
        }

        while session.status in ['starting', 'running']:
            now = time.time()
            session.data['timestamp'] = now

            ops = self.hwp_state.update(test_mode=test_mode)
            session.data['monitored_sessions'] = ops
            session.data['hwp_state'] = asdict(self.hwp_state)

            # Get actions for each hwp subsystem
            session.data['actions'] = {
                'pmx': self.hwp_state.pmx_action,
                'gripper': self.hwp_state.gripper_action,
            }

            if test_mode:
                break

            pm.sleep()

        return True, "Monitor process stopped"

    def _stop_monitor(self, session, params):
        session.status = 'stopping'
        return True, 'Stopping monitor process'

    @ocs_agent.param('test_mode', type=bool, default=False)
    def spin_control(self, session, params):
        """spin_control()

        **Process** - Process to manage the spin-state for HWP agents. This will
        issue commands to various HWP agents depending on the current control
        state.

        Args
        ----------
        test_mode : bool
            If True, spin_control loop will run a single update iteration before
            exiting. This is useful for testing actions.

        Notes
        --------

        Example of ``session.data``::

            >>> session['data']
            {
                'current_action': <Encoded action>,
                'action_history': List[Encoded action]
                'timestammp': <time.time()>
            }
        """
        clients = self._get_hwp_clients()

        def update_supervisor_state():
            state_info = self.control_state_machine.action.cur_state_info.encode()
            self.hwp_state.supervisor_control_state = state_info
            if session.data:
                session.data['current_state_type'] = self.control_state_machine.action.cur_state_info.state_type

        while session.status in ['starting', 'running']:
            # Update session data beginning and end of update call, because its
            # possible for an action change to happen due to external request from
            # task.
            update_supervisor_state()
            self.control_state_machine.update(clients, self.hwp_state)
            update_supervisor_state()
            session.data.update({
                'current_action': self.control_state_machine.action.encode(),
                'action_history': [a.encode() for a in self.control_state_machine.action_history],
                'timestamp': time.time(),
            })
            if params['test_mode']:
                break
            time.sleep(1)
        return True, "Finished spin control process"

    def _stop_spin_control(self, session, params):
        session.status = 'stopping'
        return True, 'Stopping spin control process'

    @ocs_agent.param('target_freq', type=float)
    @ocs_agent.param('freq_tol', type=float, default=0.05)
    @ocs_agent.param('freq_tol_duration', type=float, default=10)
    def pid_to_freq(self, session, params):
        """pid_to_freq(target_freq=2.0, freq_tol=0.05, freq_tol_duration=10)

        **Task** - Sets the control state to PID the HWP to the given ``target_freq``.

        Args
        -------
        target_freq : float
            Target frequency of the HWP (Hz). This is aa signed float where
            positive values correspond to counter-clockwise motion, as seen when
            looking at the cryostat from the sky.
        freq_tol : float
            Frequency threshold (Hz) for determining when the HWP is at the target frequency.
        freq_tol_duration : float
            Duration (seconds) for which the HWP must be within ``freq_tol`` of the
            ``target_freq`` to be considered successful.

        Notes
        --------

        Example of ``session.data``::

            >>> session['data']
            {'action':
                {'action_id': 3,
                'completed': True,
                'cur_state': {'class': 'Done', 'msg': None, 'success': True},
                'state_history': List[ConrolState],
                'success': True}
            }
        """
        if params['target_freq'] >= 0:
            d = '1' if self.forward_is_cw else '0'
        else:
            d = '0' if self.forward_is_cw else '1'

        state = ControlState.PIDToFreq(
            target_freq=np.abs(params['target_freq']),
            freq_tol=params['freq_tol'],
            freq_tol_duration=params['freq_tol_duration'],
            direction=d
        )
        action = self.control_state_machine.request_new_action(state)
        action.sleep_until_complete(session=session)
        return action.success, f"Completed with state: {action.cur_state_info.state}"

    @ocs_agent.param('voltage', type=float)
    @ocs_agent.param('direction', type=str, choices=['cw', 'ccw'], default='cw')
    def set_const_voltage(self, session, params):
        """set_const_voltage(voltage=1.0)

        **Task** - Sets the control state set the PMX to a constant voltage.

        Args
        -------
        voltage : float
            Voltage to set the PMX to (V).
        direction : str
            Direction of the HWP. Must be one of ``cw`` or ``ccw``,
            corresponding to the clockwise and counter-clockwise directions of
            the HWP, as seen when looking at the cryostat from the sky.

        Notes
        --------

        Example of ``session.data``::

            >>> session['data']
            {'action':
                {'action_id': 3,
                'completed': True,
                'cur_state': {'class': 'Done', 'msg': None, 'success': True},
                'state_history': List[ConrolState],
                'success': True}
            }
        """
        if params['direction'] == 'cw':
            d = '0' if self.forward_is_cw else '1'
        else:
            d = '1' if self.forward_is_cw else '0'
        state = ControlState.ConstVolt(
            voltage=params['voltage'],
            direction=d
        )
        action = self.control_state_machine.request_new_action(state)
        action.sleep_until_complete(session=session)
        return action.success, f"Completed with state: {action.cur_state_info.state}"

    @ocs_agent.param('freq_tol', type=float, default=0.05)
    @ocs_agent.param('freq_tol_duration', type=float, default=10)
    @ocs_agent.param('brake_voltage', type=float, default=10.)
    def brake(self, session, params):
        """brake(freq_tol=0.05, freq_tol_duration=10, brake_voltage=10)

        **Task** - Sets the control state to brake the HWP.

        Args
        -------
        freq_tol : float
            Frequency tolerance (Hz) for determining when the HWP is at the target frequency.
        freq_tol_duration : float
            Duration (seconds) for which the HWP must be within ``freq_tol`` of the
            ``target_freq`` to be considered successful.
        brake_voltage: float
            Voltage to use when braking the HWP.

        Notes
        --------

        Example of ``session.data``::

            >>> session['data']
            {'action':
                {'action_id': 3,
                'completed': True,
                'cur_state': {'class': 'Done', 'msg': None, 'success': True},
                'state_history': List[ControlState],
                'success': True}
            }
        """
        state = ControlState.Brake(
            freq_tol=params['freq_tol'],
            freq_tol_duration=params['freq_tol_duration'],
            brake_voltage=params['brake_voltage'],
        )
        action = self.control_state_machine.request_new_action(state)
        action.sleep_until_complete(session=session)
        return action.success, f"Completed with state: {action.cur_state_info.state}"

    @ocs_agent.param('wait_stop', type=bool, default=False)
    @ocs_agent.param('freq_tol', type=float, default=0.05)
    @ocs_agent.param('freq_tol_duration', type=float, default=30)
    def pmx_off(self, session, params):
        """pmx_off(wait_stop=False, freq_tol=0.05, freq_tol_duration=30)

        **Task** - Sets the control state to turn off the PMX.

        Args
        -------
        wait_stop : bool
            Whether to wait until hwp stops.
        freq_tol : float
            Tolerance of frequency to consider hwp is stopped.
        freq_tol_duration : float
            Duration in seconds that the frequency must be within the tolerance

        Notes
        --------

        Example of ``session.data``::

            >>> session['data']
            {'action':
                {'action_id': 3,
                'completed': True,
                'cur_state': {'class': 'Done', 'msg': None, 'success': True},
                'state_history': List[ConrolState],
                'success': True}
            }
        """
        state = ControlState.PmxOff(
            wait_stop=params['wait_stop'],
            freq_tol=params['freq_tol'],
            freq_tol_duration=params['freq_tol_duration'],
        )
        action = self.control_state_machine.request_new_action(state)
        action.sleep_until_complete(session=session)
        return action.success, f"Completed with state: {action.cur_state_info.state}"

    def grip_hwp(self, session, params):
        """grip_hwp()

        **Task** - Grip the HWP if ACU and HWP state passes checks.

        Notes
        --------

        Example of ``session.data``::

            >>> session['data']
            {'action':
                {'action_id': 3,
                'completed': True,
                'cur_state': {'class': 'Done', 'msg': None, 'success': True},
                'state_history': List[ConrolState],
                'success': True}
            }
        """
        state = ControlState.GripHWP()
        action = self.control_state_machine.request_new_action(state)
        action.sleep_until_complete(session=session)
        return action.success, f"Completed with state: {action.cur_state_info.state}"

    def ungrip_hwp(self, session, params):
        """ungrip_hwp()

        **Task** - Ungrip the HWP if ACU and HWP state passes checks.

        Notes
        --------

        Example of ``session.data``::

            >>> session['data']
            {'action':
                {'action_id': 3,
                'completed': True,
                'cur_state': {'class': 'Done', 'msg': None, 'success': True},
                'state_history': List[ConrolState],
                'success': True}
            }
        """
        state = ControlState.UngripHWP()
        action = self.control_state_machine.request_new_action(state)
        action.sleep_until_complete(session=session)
        return action.success, f"Completed with state: {action.cur_state_info.state}"

    def abort_action(self, session, params):
        """abort_action()

        **Task** - Aborts the current action, setting the control state to Idle

        Notes
        --------

        Example of ``session.data``::

            >>> session['data']
            {'action':
                {'action_id': 3,
                'completed': True,
                'cur_state': {'class': 'Idle'},
                'state_history': List[ConrolState],
                'success': False}
            }
        """
        state = ControlState.Idle()
        action = self.control_state_machine.request_new_action(state)
        session.data['action'] = action.encode()
        return True, "Set state to idle"

    def enable_driver_board(self, session, params):
        """enable_driver_board()

        **Task** - Enables the HWP driver board

        Notes
        --------

        Example of ``session.data``::

            >>> session['data']
            {'action':
                {'action_id': 3,
                'completed': True,
                'cur_state': {'class': 'Done', 'msg': None, 'success': True},
                'state_history': List[ConrolState],
                'success': False}
            }
        """
        kw = {
            'driver_power_agent_type': self.driver_power_agent_type,
            'outlets': self.driver_iboot_outlets,
            'cycle_twice': self.driver_power_cycle_twice,
            'cycle_wait_time': self.driver_power_cycle_wait_time,
        }
        state = ControlState.EnableDriverBoard(**kw)
        action = self.control_state_machine.request_new_action(state)
        action.sleep_until_complete(session=session)
        return action.success, f"Completed with state: {action.cur_state_info.state}"

    def disable_driver_board(self, session, params):
        """disable_driver_board()

        **Task** - Disables the HWP driver board

        Notes
        --------

        Example of ``session.data``::

            >>> session['data']
            {'action':
                {'action_id': 3,
                'completed': True,
                'cur_state': {'class': 'Done', 'msg': None, 'success': True},
                'state_history': List[ConrolState],
                'success': False}
            }
        """
        kw = {
            'driver_power_agent_type': self.driver_power_agent_type,
            'outlets': self.driver_iboot_outlets,
        }
        state = ControlState.DisableDriverBoard(**kw)
        action = self.control_state_machine.request_new_action(state)
        action.sleep_until_complete(session=session)
        return action.success, f"Completed with state: {action.cur_state_info.state}"


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()
    pgroup = parser.add_argument_group('Agent Options')

    pgroup.add_argument('--sleep-time', type=float, default=2.)
    pgroup.add_argument('--ybco-lakeshore-id',
                        help="Instance ID for lakeshore reading out HWP temp")
    pgroup.add_argument('--ybco-temp-field',
                        help='Field name of lakeshore channel reading out HWP temp')
    pgroup.add_argument('--ybco-temp-thresh', type=float,
                        help="Threshold for HWP temp.")
    pgroup.add_argument('--hwp-encoder-id',
                        help="Instance id for HWP encoder agent")
    pgroup.add_argument('--hwp-pmx-id',
                        help="Instance ID for HWP pmx agent")
    pgroup.add_argument('--hwp-pid-id',
                        help="Instance ID for HWP pid agent")
    pgroup.add_argument('--hwp-pcu-id',
                        help="Instance ID for HWP PCU agent")
    pgroup.add_argument('--hwp-gripper-id',
                        help="Instance ID for HWP Gripper agent")
    pgroup.add_argument('--ups-id', help="Instance ID for UPS agent")
    pgroup.add_argument('--ups-minutes-remaining-thresh', type=float,
                        help="Threshold for UPS minutes remaining before a "
                             "shutdown is triggered")
    pgroup.add_argument(
        '--pid-max-time-since-update', type=float, default=60.0,
        help="Max amount of time since last PID update before data is considered stale.",
    )

    pgroup.add_argument(
        '--driver-iboot-id',
        help="Instance ID for IBoot-PDU agent that powers the HWP Driver board")
    pgroup.add_argument(
        '--driver-iboot-outlets', nargs='+', type=int,
        help="Outlets for driver iboot power")
    pgroup.add_argument(
        '--driver-power-cycle-twice', action='store_true',
        help="If set, will power cycle the driver board twice on enable")
    pgroup.add_argument(
        '--driver-power-cycle-wait-time', type=float, default=60 * 5,
        help="Wait time between power cycles on enable (sec)")
    pgroup.add_argument(
        '--driver-power-agent-type', choices=['iboot', 'synaccess'], default=None,
        help="Type of agent used for controlling the driver power")

    pgroup.add_argument(
        '--gripper-iboot-id',
        help="Instance ID for IBoot-PDU agent that powers the gripper controller")
    pgroup.add_argument(
        '--gripper-iboot-outlets', nargs='+', type=int,
        help="Outlets for gripper iboot power")
    pgroup.add_argument(
        '--gripper-power-agent-type', choices=['iboot', 'synaccess'], default=None,
        help="Type of agent used for controlling the gripper power")

    pgroup.add_argument(
        '--acu-instance-id', default='acu',
        help="Instance ID for the ACU agent."
    )
    pgroup.add_argument(
        '--no-acu', action='store_true',
        help="If set, will not attempt to connect to the ACU or perform any ACU "
             "checks for grip and spin-up"
    )
    pgroup.add_argument(
        '--acu-min-el', type=float, default=48.0,
        help="Min elevation that HWP spin up is allowed",
    )
    pgroup.add_argument(
        '--acu-max-el', type=float, default=90.0,
        help="Max elevation that HWP spin up is allowed",
    )
    pgroup.add_argument(
        '--acu-tol-el', type=float, default=0.1,
        help="Amount of tolerance in elevation allowed when HWP spin up",
    )
    pgroup.add_argument(
        '--acu-max-time-since-update', type=float, default=30.0,
        help="Max amount of time since last ACU update before allowing HWP spin up",
    )
    pgroup.add_argument(
        '--mount-velocity-grip-thresh', type=float, default=0.005,
        help="Max mount velocity (both az and el) for gripping the HWP"
    )
    pgroup.add_argument(
        '--grip-max-boresight-angle', type=float, default=1.,
        help="Maximum absolute value of boresight angle (deg) for gripping the HWP"
    )
    pgroup.add_argument(
        '--use-acu-blocking', action='store_true',
        help="If True, will use blocking flags to make sure ACU is not moving "
             "during HWP gripper commands."
    )
    pgroup.add_argument(
        '--acu-block-motion-timeout', type=float, default=60.,
        help="Time to wait for ACU motion to be blocked before aborting gripper command."
    )
    pgroup.add_argument(
        '--forward-dir', choices=['cw', 'ccw'], default='cw',
        help="Whether the PID 'forward' direction is cw or ccw. "
             "cw or ccw is defined when hwp is viewed from the sky side."
    )
    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='HWPSupervisor',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)
    hwp = HWPSupervisor(agent, args)

    agent.register_process('monitor', hwp.monitor, hwp._stop_monitor,
                           startup=True)
    agent.register_process(
        'spin_control', hwp.spin_control, hwp._stop_spin_control, startup=True)
    agent.register_task('query_hwp_state', hwp.query_hwp_state)
    agent.register_task('pid_to_freq', hwp.pid_to_freq)
    agent.register_task('set_const_voltage', hwp.set_const_voltage)
    agent.register_task('brake', hwp.brake)
    agent.register_task('pmx_off', hwp.pmx_off)
    agent.register_task('grip_hwp', hwp.grip_hwp)
    agent.register_task('ungrip_hwp', hwp.ungrip_hwp)
    agent.register_task('enable_driver_board', hwp.enable_driver_board)
    agent.register_task('disable_driver_board', hwp.disable_driver_board)
    agent.register_task('abort_action', hwp.abort_action)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import pickle as pkl
import socket
import time


class GripperClient(object):
    def __init__(self, ip, port):
        self.ip = ip
        self.port = port
        self.control_socket = self._init_connection(self.ip, self.port)
        self.data = b''

    def _init_connection(self, ip, port):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((ip, port))
        return s

    def power(self, state):
        if not isinstance(state, bool):
            return False

        if state:
            return self.send_command('ON')
        else:
            return self.send_command('OFF')

    def home(self):
        return self.send_command('HOME')

    def move(self, mode, actuator, distance):
        if not isinstance(mode, str):
            return False
        elif mode.upper() not in ['POS', 'PUSH']:
            return False

        if actuator not in [1, 2, 3]:
            return False

        if type(distance) not in [float, int]:
            return False

        return self.send_command('MOVE ' + mode + ' ' + str(actuator) + ' ' + str(distance))

    def brake(self, state, actuator=0):
        if not isinstance(state, bool):
            return False

        if actuator == 0:
            if state:
                return self.send_command('BRAKE ON')
            else:
                return self.send_command('BRAKE OFF')
        else:
            if actuator not in [1, 2, 3]:
                return False

            if state:
                return self.send_command('BRAKE ON ' + str(actuator))
            else:
                return self.send_command('BRAKE OFF ' + str(actuator))

    def emg(self, state, actuator=0):
        if not isinstance(state, bool):
            return False

        if actuator == 0:
            if state:
                return self.send_command('EMG ON')
            else:
                return self.send_command('EMG OFF')
        else:
            if actuator not in [1, 2, 3]:
                return False

            if state:
                return self.send_command('EMG ON ' + str(actuator))
            else:
                return self.send_command('EMG OFF ' + str(actuator))

    def alarm(self):
        return self.send_command('ALARM')

    def alarm_group(self):
        return self.send_command('ALARM_GROUP')

    def reset(self):
        return self.send_command('RESET')

    def inp(self):
        return self.send_command('INP')

    def act(self, actuator):
        if actuator not in [1, 2, 3]:
            return False

        return self.send_command('ACT ' + str(actuator))

    def is_cold(self, value):
        if value:
            return self.send_command('IS_COLD 1')
        else:
            return self.send_command('IS_COLD 0')

    def force(self, value):
        if value:
            return self.send_command('FORCE 1')
        else:
            return self.send_command('FORCE 0')

    def get_state(self):
        return self.send_command('GET_STATE')

    def send_command(self, command):
        if isinstance(command, str):
            self.control_socket.sendall(bytes(command, 'utf-8'))
        elif isinstance(command, bytes):
            self.control_socket.sendall(command)

        return pkl.loads(self.control_socket.recv(4096))

    def restart(self):
        log = []
        try:
            self.close()
            log.append('Previous connection closed')
        except BaseException:
            log.append('Previous connection already closed')

        _restart_socket = self._init_connection(self.ip, 5656)
        _restart_socket.sendall(('reset\n').encode())
        time.sleep(0.5)

        resp = _restart_socket.recv(4096).decode().strip()
        log.append(f'Restart command response: {resp}')
        result = True if resp == 'Success' else False
        _restart_socket.close()
        time.sleep(10)

        self.control_socket = self._init_connection(self.ip, self.port)
        log.append('Control socket reconnected')
        return {'result': result, 'log': log}

    def close(self):
        """Close socket connection"""
        self.control_socket.close()



# Tucker Elleflot

import socket
import time

from socs.common.prologix_interface import PrologixInterface

# append new model strings as needed
ONE_CHANNEL_MODELS = ['2280S-60-3', '2280S-32-6', '9171',
                      '9172', '9181', '9182', '9183', '9184', '9185']
TWO_CHANNEL_MODELS = ['9173', '9174']
THREE_CHANNEL_MODELS = ['2230G-30-1']

# error codes from 2280S devices
# part of an attempt to query devices for the number
# of supported channels
UNDEFINED_HEADER = -113
HEADER_SUFFIX_OUT_OF_RANGE = -114


class ScpiPsuInterface:
    def __init__(self, ip_address, port, **kwargs):
        self.ip_address = ip_address
        self.port = port
        self.sock = None
        self.model = None
        self.num_channels = 0
        self.conn_socket()
        try:
            self.configure()
        except ValueError as err:
            raise ValueError(err)

    def conn_socket(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect((self.ip_address, self.port))
        self.sock.settimeout(5)

    def read(self):
        return self.sock.recv(128).decode().strip()

    def write(self, msg):
        message = msg + '\n'
        self.sock.sendall(message.encode())
        time.sleep(0.1)  # to prevent flooding the connection

    def identify(self):
        self.write('*idn?')
        return self.read()

    def read_model(self):
        idn_response = self.identify().split(',')[1]
        if (idn_response.startswith('MODEL')):
            return idn_response[6:]
        else:
            return idn_response

    def configure(self):
        self.model = self.read_model()

        if (self.model in ONE_CHANNEL_MODELS):
            self.num_channels = 1
        if (self.model in TWO_CHANNEL_MODELS):
            self.num_channels = 2
        if (self.model in THREE_CHANNEL_MODELS):
            self.num_channels = 3
        if (self.num_channels == 0):
            self.num_channels = 3
            raise ValueError('Model number not found in known device models', self.model)

    def enable(self, ch):
        '''
        Enables output for channel (1,2,3) but does not turn it on.
        Depending on state of power supply, it might need to be called
        before the output is set.
        '''
        if (self.num_channels != 1):
            self.set_chan(ch)
        self.write('OUTP:ENAB ON')

    def disable(self, ch):
        '''
        disabled output from a channel (1,2,3). once called, enable must be
        called to turn on the channel again
        '''
        self.write('OUTP:ENAB OFF')

    def set_chan(self, ch):
        if (self.num_channels != 1):
            self.write('inst:nsel ' + str(ch))

    def set_output(self, ch, out):
        '''
        set status of power supply channel
        ch - channel (1,2,3) to set status
        out - ON: True|1|'ON' OFF: False|0|'OFF'

        Calls enable to ensure a channel can be turned on. We might want to
        make them separate (and let us use disable as a safety feature) but
        for now I am thinking we just want to thing to turn on when we tell
        it to turn on.
        '''
        self.set_chan(ch)
        self.enable(ch)
        if (self.num_channels != 1):
            if isinstance(out, str):
                self.write('CHAN:OUTP ' + out)
            elif out:
                self.write('CHAN:OUTP ON')
            else:
                self.write('CHAN:OUTP OFF')
        else:
            if isinstance(out, str):
                self.write('OUTP ' + out)
            elif out:
                self.write('OUTP ON')
            else:
                self.write('OUTP OFF')

    def get_output(self, ch):
        '''
        check if the output of a channel (1,2,3) is on (True) or off (False)
        '''
        self.set_chan(ch)
        if (self.num_channels != 1):
            self.write('CHAN:OUTP:STAT?')
        else:
            self.write('OUTP:STAT?')
        out = bool(float(self.read()))
        return out

    def set_volt(self, ch, volt):
        self.set_chan(ch)
        self.write('volt ' + str(volt))

    def set_curr(self, ch, curr):
        self.set_chan(ch)
        self.write('curr ' + str(curr))

    def get_volt(self, ch):
        self.set_chan(ch)
        if (self.num_channels != 1):
            self.write('MEAS:VOLT? CH' + str(ch))
        else:
            self.write('MEAS:VOLT?')
        voltage = float(self.read().split(',')[1].strip('V'))
        return voltage

    def get_curr(self, ch):
        self.set_chan(ch)
        if (self.num_channels != 1):
            self.write('MEAS:CURR? CH' + str(ch))
        else:
            self.write('MEAS:CURR?')
        current = float(self.read().split(',')[0].strip('A'))
        return current

    def clear(self):
        return True


class PsuInterface(PrologixInterface):
    def __init__(self, ip_address, gpibAddr, verbose=False, **kwargs):
        self.verbose = verbose
        self.num_channels = 0
        super().__init__(ip_address, gpibAddr, **kwargs)

    def enable(self, ch):
        '''
        Enables output for channel (1,2,3) but does not turn it on.
        Depending on state of power supply, it might need to be called
        before the output is set.
        '''
        self.set_chan(ch)
        self.write('OUTP:ENAB ON')

    def disable(self, ch):
        '''
        disabled output from a channel (1,2,3). once called, enable must be
        called to turn on the channel again
        '''
        self.write('OUTP:ENAB OFF')

    def set_chan(self, ch):
        self.write('inst:nsel ' + str(ch))

    def set_output(self, ch, out):
        '''
        set status of power supply channel
        ch - channel (1,2,3) to set status
        out - ON: True|1|'ON' OFF: False|0|'OFF'

        Calls enable to ensure a channel can be turned on. We might want to
        make them separate (and let us use disable as a safety feature) but
        for now I am thinking we just want to thing to turn on when we tell
        it to turn on.
        '''
        self.set_chan(ch)
        self.enable(ch)
        if isinstance(out, str):
            self.write('CHAN:OUTP ' + out)
        elif out:
            self.write('CHAN:OUTP ON')
        else:
            self.write('CHAN:OUTP OFF')

    def get_output(self, ch):
        '''
        check if the output of a channel (1,2,3) is on (True) or off (False)
        '''
        self.set_chan(ch)
        self.write('CHAN:OUTP:STAT?')
        out = bool(float(self.read()))
        return out

    def set_volt(self, ch, volt):
        self.set_chan(ch)
        self.write('volt ' + str(volt))
        if self.verbose:
            voltage = self.get_volt(ch)
            print("CH " + str(ch) + " is set to " + str(voltage) + " V")

    def set_curr(self, ch, curr):
        self.set_chan(ch)
        self.write('curr ' + str(curr))
        if self.verbose:
            current = self.get_curr(ch)
            print("CH " + str(ch) + " is set to " + str(current) + " A")

    def get_volt(self, ch):
        self.set_chan(ch)
        self.write('MEAS:VOLT? CH' + str(ch))
        voltage = float(self.read())
        return voltage

    def get_curr(self, ch):
        self.set_chan(ch)
        self.write('MEAS:CURR? CH' + str(ch))
        current = float(self.read())
        return current

    def clear(self):
        # Clear all the event registers and error queue, using a query such as *ESR? or MEAS:X?
        # instead of *CLS can confuse the PSU
        self.write('*CLS')
        return True



#!/usr/bin/env python3
"""Module to handle thermometers connected to KR260 via SPI bus.
"""

from pathlib import Path


class StimThermoError(Exception):
    """Exception raised by stimulator thermometer."""


class Iio:
    """Generic IIO driver.

    Parameters
    ----------
    path : str or pathlib.Path
        Path to the iio device directory.
    """

    def __init__(self, path):
        self._path = Path(path)

    def read(self, name):
        """Read the file in the iio directory.

        Parameters
        ----------
        name : str
            File name in IIO directory.

        Returns
        -------
        data : str
            Contents of the file.
        """
        with open(self._path.joinpath(name), 'r') as fd:
            return fd.readline().strip()

    @property
    def name(self):
        """Read name of the iio device.

        Returns
        -------
        name : str
            Name of the device.
        """
        return self.read('name')


class Max31856(Iio):
    """Class to read MAX31856 via IIO interface.

    Parameters
    ----------
    path : str or pathlib.Path
        Path to the iio device directory.
    """
    @property
    def thermocouple_type(self):
        """Get thermocouple type.

        Returns
        -------
        type : str
            Thermocouple type in string.
        """
        return self.read('in_temp_thermocouple_type')

    def get_temp_raw(self):
        """Get raw temperature.

        Returns
        -------
        temp_raw : int
            Raw temperature.
        """
        return int(self.read('in_temp_raw'))

    def get_temp(self):
        """Get temperature in degrees Celsius.

        Returns
        -------
        temp : float
            Temperature in degrees Celsius.
        """
        return self.get_temp_raw() / 2**7

    def get_temp_ambient_raw(self):
        """Get raw ambient temperature.

        Returns
        -------
        temp_ambient_raw : int
            Raw ambient temperature.
        """
        return int(self.read('in_temp_ambient_raw'))

    def get_temp_ambient(self):
        """Get ambient temperature in degrees Celsius.

        Returns
        -------
        temp_ambient : float
            Ambient temperature in degrees Celsius.
        """
        return self.get_temp_ambient_raw() / 2**6


class Max31865(Iio):
    """Class to read MAX31865 via IIO interface.

    Parameters
    ----------
    path : str or pathlib.Path
        Path to the iio device directory.
    """
    R_REF = 4300
    R_0 = 1000
    ALPHA = 0.0039

    def get_temp_raw(self):
        """Get raw temperature.

        Returns
        -------
        temp_raw : int
            Raw temperature.
        """
        return int(self.read('in_temp_raw'))

    def get_temp(self):
        """Get temperature in degrees Celsius.

        Returns
        -------
        temp : float
            Temperature in degrees Celsius.
        """
        r_rtd = self.R_REF * self.get_temp_raw() / 32768

        return (r_rtd / self.R_0 - 1) / self.ALPHA


def from_spi_node_path(path):
    """Return device class from path to the SPI directory.

    Paramter
    --------
    path : str or pathlib.Path
        Path to the SPI node directory like
        /sys/bus/spi/devices/spi3.0/

    Returns
    -------
    dev : Max31865 or Max31856
        Device class.
    """
    path_iio = list(Path(path).glob('iio*'))
    if len(path_iio) != 1:
        raise StimThermoError(f'IIO definition not found: {path_iio}')

    dev_iio = Iio(path_iio[0])

    name_dev = dev_iio.name

    if name_dev == 'max31856':
        return Max31856(path_iio[0])
    elif name_dev == 'max31865':
        return Max31865(path_iio[0])
    else:
        raise StimThermoError('Thermometer type not supported.')


def main():
    """ Main function """
    paths = ['/sys/bus/spi/devices/spi3.0/',
             '/sys/bus/spi/devices/spi3.1/',
             '/sys/bus/spi/devices/spi3.3/',
             '/sys/bus/spi/devices/spi3.4/']
    thermometers = [from_spi_node_path(path) for path in paths]

    for dev in thermometers:
        print(dev.get_temp())
        if isinstance(dev, Max31856):
            print(dev.get_temp_ambient())


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
'''PCR500MA driver using SCPI-RAW protocol'''
import sys
from enum import Enum

from socs.tcp import TCPInterface

PORT = 5025
TIMEOUT_DEFAULT = 3
BUFFSIZE = 4096

# software limits
VOLT_ULIM_SOFT = 51


class PCRException(Exception):
    pass


class PCRCoupling(Enum):
    AC = 'AC'
    DC = 'DC'
    ACDC = 'ACDC'


class PCR500MA(TCPInterface):
    """PCR500MA driver

    Parameters
    ----------
    ip_addr : str
        IP address of the device
    port : int
        Port number
    timeout : float
        Timeout count in seconds
    """

    def __init__(self, ip_addr, port=PORT, timeout=TIMEOUT_DEFAULT):
        super().__init__(ip_addr, port, timeout)

    def _send(self, message):
        """Send message to PCR500MA

        Parameters
        ----------
        message : str
            Message string to be sent to PCR500MA
        """
        if not message[-1] == '\n':
            message += '\n'

        self.send(message.encode())

    def _recv(self):
        """Receive message from PCR500MA"""
        msg_byte = self.recv(BUFFSIZE)

        return msg_byte.decode()

    def _query(self, message):
        self._send(message)
        return self._recv().strip()

    def get_idn(self):
        """Get identification string"""
        return self._query('*IDN?')

    def reset(self):
        """Reset PCR500MA device"""
        return self._send('*RST')

    def set_output(self, output):
        """Set output status

        Parameters
        ----------
        output : bool
            True : output on
            False : output off
        """
        message = f'OUTP {"1" if output else "0"}'
        self._send(message)

    def turn_on(self):
        """Turn on"""
        self.set_output(True)

    def turn_off(self):
        """Trun off"""
        self.set_output(False)

    def get_output(self):
        """Get output status

        Returns
        -------
        status : bool
            True : output on
            False : output off
        """
        message = self._query('OUTP?')

        if message == '1':
            return True
        elif message == '0':
            return False
        else:
            raise PCRException(f'Failed to get output status: {message}')

    def set_coupling(self, pcr_coupling):
        """Set output coupling

        Parameters
        ----------
        pcr_coupling : PCRCoupling
            PCR coupling
        """
        message = f'OUTP:COUP {pcr_coupling.value}'
        self._send(message)

    def get_coupling(self):
        """Get output coupling

        Returns
        -------
        pcr_coupling : PCRCoupling
            PCR coupling
        """
        message = self._query('OUTP:COUP?')

        return PCRCoupling(message)

    def clear_alarm(self):
        """Clear alarm"""
        self._send('OUTP:PROT:CLE')

    def set_current_limit_ac(self, current_amp):
        """Set AC current upper limit

        Parameters
        ----------
        current_amp : float
            AC currrent upper limit in Ampere.
        """
        message = f'CURR {current_amp}'
        self._send(message)

    def get_current_limit_ac(self):
        """Get AC current upper limit

        Returns
        -------
        current_amp : float
            AC current upper limit in Ampere.
        """
        message = self._query('CURR?')
        return float(message)

    def set_current_limit_dc(self, current_amp):
        """Set DC current upper limit

        Parameters
        ----------
        current_amp : float
            DC currrent upper limit in Ampere.
        """
        message = f'CURR:OFFS {current_amp}'
        self._send(message)

    def get_current_limit_dc(self):
        """Get DC current upper limit

        Returns
        -------
        current_amp : float
            DC current upper limit in Ampere.
        """
        message = self._query('CURR:OFFS?')
        return float(message)

    def set_frequency(self, freq, freq_llim=None, freq_ulim=None):
        """Set AC frequency.

        Parameters
        ----------
        freq : float
            Frequency in Hz.
        freq_llim : float
            Frequency lower limit in Hz.
        freq_ulim : float
            Frequency upper limit in Hz.
        """

        assert 40 <= freq <= 500, 'Frequency should be within [40, 500] Hz'

        if (freq_llim is None) and (freq_ulim is None):
            message = f'FREQ {freq}'
        elif (freq_llim is None) or (freq_ulim is None):
            raise PCRException('Lower limit and upper limit should be given simultaneously.')
        else:
            assert freq_llim <= freq <= freq_ulim, 'Target frequency is not in the given range'
            message = f'FREQ {freq},{freq_llim},{freq_ulim}'

        self._send(message)

    def get_frequency(self):
        """Get AC frequency

        Returns
        -------
        freq : float
            Frequency in Hz
        """

        message = self._query('FREQ?')
        return float(message)

    def set_volt_ac(self, volt, volt_llim=None, volt_ulim=None):
        """Set AC voltage

        Parameters
        ----------
        volt : float
            AC voltage in V.
        volt_llim : float
            AC voltage lower limit in V.
        volt_ulim : float
            AC voltage upper limit in V.
        """
        assert 0 <= volt <= VOLT_ULIM_SOFT, f'Voltage should be within [0, {VOLT_ULIM_SOFT}] V'

        if (volt_llim is None) and (volt_ulim is None):
            message = f'VOLT {volt}'
        elif (volt_llim is None) or (volt_ulim is None):
            raise PCRException('Lower limit and upper limit should be given simultaneously.')
        else:
            assert volt_llim <= volt <= volt_ulim, 'Target voltage is not in the given range'
            message = f'VOLT {volt},{volt_llim},{volt_ulim}'

        self._send(message)

    def get_volt_ac(self):
        """Get AC voltage setting

        Returns
        -------
        volt : float
            AC voltage setting in V.
        """

        message = self._query('VOLT?')
        return float(message)

    def meas_volt_ac(self):
        """Measure AC voltage

        Returns
        -------
        volt : float
            Measured AC voltage in V.
        """

        message = self._query('MEAS:VOLT:AC?')
        return float(message)

    def meas_current_ac(self):
        """Measure AC current

        Returns
        -------
        current : float
            Measured AC current in A.
        """

        message = self._query('MEAS:CURR:AC?')
        return float(message)

    def meas_power_ac(self):
        """Measure AC power

        Returns
        -------
        power : float
            Measured power in W.
        """
        message = self._query('MEAS:POW:AC?')
        return float(message)

    def meas_freq(self):
        """Measure frequency

        Returns
        -------
        freq : float
            Measured frequency in Hz.
        """
        message = self._query('MEAS:FREQ?')
        return float(message)


def main():
    """Main function"""
    inst = PCR500MA(sys.argv[1])
    print(inst.get_idn())
    print(inst.get_output())
    print(inst.get_coupling())
    print(inst.get_current_limit_ac())
    print(inst.get_current_limit_dc())
    print(inst.get_frequency())
    print(inst.get_volt_ac())
    print(inst.meas_volt_ac())
    print(inst.meas_current_ac())
    print(inst.meas_power_ac())
    print(inst.meas_freq())


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
Utility script for interacting with the suprsync db.
"""
import argparse
import os
import time

from tqdm.auto import trange

from socs.db.suprsync import SupRsyncFile, SupRsyncFilesManager


def check_func(args):
    srfm = SupRsyncFilesManager(args.db, create_all=False)
    session = srfm.Session()
    files = session.query(SupRsyncFile).filter(
        SupRsyncFile.local_path == args.file
    ).all()
    for file in files:
        print('-' * 80)
        print(file)
        print('-' * 80)


def next_func(args):
    srfm = SupRsyncFilesManager(args.db, create_all=False)

    files = srfm.get_copyable_files(args.archive_name, num_files=1)
    if not files:
        print("No more files left to be copied")
    else:
        print(files[0])


def list_func(args):
    srfm = SupRsyncFilesManager(args.db, create_all=False)
    files = srfm.get_copyable_files(args.archive_name)
    if not files:
        print("No files left to be copied")
        return

    for i, f in enumerate(files):
        print(f"{i}: {f.local_path}")


def add_local_files_func(args):
    srfm = SupRsyncFilesManager(args.db, create_all=args.create_db)

    args.local_root = os.path.abspath(args.local_root)

    known_files = srfm.get_known_files(args.archive_name)
    known_paths = [f.local_path for f in known_files]
    local_paths = []
    remote_paths = []
    now = time.time()

    for root, _, files in os.walk(args.local_root):
        for file in files:
            path = os.path.join(root, file)
            path = os.path.abspath(path)
            if path not in known_paths and now - os.stat(path).st_mtime >= args.last_edit:
                local_paths.append(path)
                remote_paths.append(os.path.relpath(path, args.local_root))

    if args.dry:
        print("Dry run\n" + 40 * '-')

        if len(local_paths) == 0:
            print("No files to add")
            return

        print(f"Would add {len(local_paths)} files, including:")
        n = min(len(local_paths), 10)
        for i in range(n):
            print(f"{i}: {local_paths[i]} --> {remote_paths[i]}")
        return

    print(f"Adding {len(local_paths)} files to the add to {args.db} from {args.local_root}")
    for i in trange(len(local_paths)):
        srfm.add_file(local_paths[i], remote_paths[i], args.archive_name)


def main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    default_db = os.environ.get("SUPRSYNC_DB")

    check_parser = subparsers.add_parser(
        'check', help="Checks suprsync file entry for a given local file.")
    check_parser.set_defaults(func=check_func)
    check_parser.add_argument('file', type=str,
                              help="local path to check in the suprsync db")
    check_parser.add_argument('--db', default=default_db, help='db path')

    next_parser = subparsers.add_parser(
        'next',
        help="Returns file info for next file to be handled in the db"
    )
    next_parser.set_defaults(func=next_func)
    next_parser.add_argument('archive_name',
                             help='Name of archive to get next file')
    next_parser.add_argument('--db', default=default_db, help='db path')

    list_parser = subparsers.add_parser('list')
    list_parser.set_defaults(func=list_func)
    list_parser.add_argument('archive_name',
                             help='Name of archive to get next file')
    list_parser.add_argument('--db', default=default_db)

    add_local_files_parser = subparsers.add_parser(
        'add-local-files',
        help="Adds all files from a local directory to the suprsync db"
    )
    add_local_files_parser.set_defaults(func=add_local_files_func)
    add_local_files_parser.add_argument(
        'local_root', help="Root directory to search for files")
    add_local_files_parser.add_argument(
        'archive_name', help="Archive to add files to")
    add_local_files_parser.add_argument('--db', default=default_db)
    add_local_files_parser.add_argument('--last-edit', default=60,
                                        help="Only add files that were last edited more than some seconds ago. Default:60")
    add_local_files_parser.add_argument(
        '--create-db', action='store_true',
        help="Create the db if it doesn't exist"
    )
    add_local_files_parser.add_argument('--dry', action='store_true',
                                        help="Does a dry run and prints what files would be added")

    args = parser.parse_args()

    if not hasattr(args, 'func'):
        parser.print_help()
        return

    if args.db is None:
        raise FileNotFoundError("Suprsync DB must be specified either through the --db "
                                "cli-arg or the SUPRSYNC_DB env var")

    args.func(args)


if __name__ == '__main__':
    main()



import selectors
import socket


class TCPInterface:
    """Interface class for connecting to devices using TCP.

    Parameters
    ----------
    ip_address : str
        IP address of the device.
    port : int
        Associated port for TCP communication.
    timeout : float
        Duration in seconds that operations wait before giving up.

    Attributes
    ----------
    ip_address : str
        IP address of the device.
    port : int
        Associated port for TCP communication.
    timeout : float
        Duration in seconds that operations wait before giving up.
    comm : socket.socket
        Socket object that forms the connection to the device.

    """

    def __init__(self, ip_address, port, timeout):
        self.ip_address = ip_address
        self.port = int(port)
        self.timeout = timeout
        self.comm = self._connect((self.ip_address, self.port))

    def _connect(self, address):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(self.timeout)
        try:
            sock.connect(address)
        except TimeoutError:
            print(f"Connection not established within {self.timeout}.")
            return
        except OSError as e:
            print(f"Unable to connect. {e}")
            return
        except Exception as e:
            print(f"Caught unexpected {type(e).__name__} while connecting:")
            print(f"  {e}")
            return
        return sock

    def _reset(self):
        print("Resetting the connection to the device.")
        self.comm = self._connect((self.ip_address, self.port))

    def send(self, msg):
        """Send message to socket.

        This method will try to send the message and if it runs into any issues
        will try to re-establish the socket connection before trying to send
        the message again. If it fails a second time it raises an exception.

        If the connection has failed to reset from a previous ``send``, or has
        not yet been established, it will first try to connnect before sending
        the message. If it fails to establish the connection it will raise an
        exception.

        Parameters
        ----------
        msg : str
            Message string to send on socket.

        Raises
        ------
        ConnectionError
            Raised if the communication fails for any reason.

        """
        if self.comm is None:
            print("Connection not established.")
            self._reset()
            if self.comm is None:
                raise ConnectionError("Unable to establish connection.")

        try:
            self.comm.sendall(msg)
            return
        except (BrokenPipeError, ConnectionResetError) as e:
            print(f"Connection error: {e}")
            self._reset()
        except TimeoutError as e:
            print(f"Timeout error while writing: {e}")
            self._reset()
        except Exception as e:
            print(f"Caught unexpected {type(e).__name__} during send:")
            print(f"  {e}")
            self._reset()

        # Try a second time before giving up
        try:
            self.comm.sendall(msg)
        except (BrokenPipeError, ConnectionResetError) as e:
            print(f"Connection error: {e}")
            raise ConnectionError
        except TimeoutError as e:
            print(f"Timeout error while writing: {e}")
            raise ConnectionError
        except AttributeError:
            raise ConnectionError("Unable to reset connection.")
        except Exception as e:
            print(f"Caught unexpected {type(e).__name__} during send:")
            print(f"  {e}")
            raise ConnectionError

    def _check_ready(self):
        """Check socket is ready to read from."""
        if self.comm is None:
            raise ConnectionError("Connection not established, not ready to read.")

        sel = selectors.DefaultSelector()
        sel.register(self.comm, selectors.EVENT_READ)
        if not sel.select(self.timeout):
            raise ConnectionError("Socket not ready to read. Possible timeout.")

    def recv(self, bufsize=4096):
        """Receive response from the device.

        This method will check if the socket is ready to be read from before
        performing the recv. If there is no data to read, or the socket is
        otherwise unready an exception is raised.

        Parameters
        ----------
        bufsize : int
            Amount of data to be recieved in bytes. Defaults to 4096.

        Returns
        -------
        ``str`` or ``bytes``
            The response from the device. The return type
            depends on the device.

        Raises
        ------
        ConnectionError
            Raised if the socket is not ready to read from.

        """
        self._check_ready()
        data = self.comm.recv(bufsize)
        return data

    def __del__(self):
        if self.comm:
            self.comm.close()
